
/*****************************************************/
/* basics.h   bliksem is a resolution based          */
/* theorem prover.                                   */
/*****************************************************/

#include <stdio.h>
#include <setjmp.h>
/* your compiler may want to see one of setjmp/setjump.h here */
#include <string.h>
#include <stdlib.h>


/********************************************************************/
/*                The name table                                    */
/********************************************************************/


/* contains the names. The names are separated by \0, freenamespace  */
/* is the index of the first free position.                          */

#define MAXNAMESPACE  1000000
CLASS char namespace [ MAXNAMESPACE ];
CLASS int freenamespace;


/* Indices into the name table. Either an index, or -1.   */

#define MAXHASHVALUE    200
CLASS int hashindex [ MAXHASHVALUE ];


#define MAXNAMETABLE  200000
CLASS int freenametable;
CLASS struct
{
   int string;  /* Reference into namespace to the string of the name. */
   int next;    /* Reference into nametable to the next name with the  */
                /* same hash value, or -1 if none exists.              */
}
   nametable [ MAXNAMETABLE ];




/*******************************************************************/
/*    Structures for the maintenance of terms:                     */
/*******************************************************************/


#define MAXOPERATORS       100000 
CLASS int arities [ MAXOPERATORS ];

#define FLAG_USER         (0X1)
   /* the operator comes from the input. */ 
#define FLAG_SKOLEM       (0X2)
   /* the operator is a Skolem function. */ 
#define FLAG_SUBFORM      (0X4) 
   /* the operator is produced by replacement of a subformula. */ 
#define FLAG_SPLITTING    (0X8) 
   /* the operator is produced by the splitting rule.          */
#define FLAG_VARIABLE     (0X10)
   /* Used as a fresh variable.                                */
#define FLAG_WORLD        (0X20) 
   /* Used to represent a world.                               */
#define FLAG_ACCESS       (0X40) 
   /* Used as accessibility relation.                          */
#define FLAG_ANSWER       (0X80)
   /* the operator is an answer literal.                       */
   /* this is usually combined with FLAG_USER.                 */

CLASS struct
{
   int name;        /* Reference into nametable.              */
   int flags;       /* See a few lines above for the values.  */ 
}
   operators [ MAXOPERATORS ];

CLASS int freeoperators; 


/* < OP_NONPRIMITIVE, operators[]. name cannot be used.   */

CLASS int freetermspace, maxtermspace, maxtermspace10;
CLASS int *termspace; 
CLASS int *termends;


#define MAXARITY 16

/* The only use of this flag is when computing a multisetextension */
/* of the recursive path order. If you don't use RPO then the      */
/* value does not matter.                                          */

/* special operators:        */ 

#define OP_NIL                  0
#define OP_DOT                  1

#define OP_FALSE                2
#define OP_TRUE                 3
#define OP_NOT                  4
#define OP_AND                  5
#define OP_OR                   6
#define OP_IMPLIES              7
#define OP_EQUIV                8

#define OP_FORALL               9
#define OP_EXISTS              10
#define OP_NECESSARY           11
#define OP_POSSIBLE            12


#define OP_EQUAL               13
#define OP_REWRITES            14

#define OP_LESS                        15
#define OP_GREATER                     16
#define OP_EQUALEQUAL                  17
#define OP_LESSEQUAL                   18
#define OP_GREATEREQUAL                19
#define OP_GREATERGREATER              20
#define OP_LESSLESS                    21
#define OP_GREATERGREATEREQUAL         22
#define OP_LESSLESSEQUAL               23
#define OP_GREATERGREATEREQUALEQUAL    24
#define OP_LESSLESSEQUALEQUAL          25
#define OP_IN                          26

#define OP_PLUS      27
#define OP_MINUS     28
#define OP_TIMES     29
#define OP_DIVIDES   30
#define OP_POW       31

#define OP_UNARYMINUS   32
#define OP_UNARYPLUS    33


/* the first non-primitive operator:  */ 
/* For primitive operators, the entry */
/* operators []. name is useless.     */ 

#define OP_NONPRIMITIVE     34



/*********************************************************/
/* structures for files:                                 */
/*********************************************************/

#define MAXFILENAMELEN  100 

#define FILE_NOTINUSE     0 
#define FILE_READ         101
#define FILE_WRITE        202

/* Used to cut off output lines, s.t. output can be printed and edited. */
#define MAXLINEPOS      75

typedef struct 
{
   FILE *handle;
   char name [ MAXFILENAMELEN ];
   int rw;    /* opened for reading, or writing?   */ 
   int linenum;
   int linepos;
   int lookahead;  
}
   bliksemfile;

CLASS bliksemfile inputfile, outputfile; 

/* Used by the Readfile command: */ 

#define MAXINCLUDEFILES  10
CLASS int freeincludefiles;
CLASS bliksemfile includefiles [ MAXINCLUDEFILES ];

/* Used by Save command:         */

CLASS bliksemfile savefile; 

CLASS int scanattr;

#define MAXNAMELEN 100


/* If the flag prologoutput is set, all output is in prolog-readable format.   */
/* We have the following collection of values:                                 */
/*    0: Standard Bliksem output.                                              */
/*    1: Prolog with standard variables.                                       */
/*    2: Prolog with variables of the form vvvv_(i).                           */
/*    3: Prolog with variables of the form vvvv_i.                             */


CLASS int prologoutput;


/**********************************************/
/*   Definitions of the types that can be     */
/*   represented by the pseudo terms.         */
/**********************************************/

/* Basic types:   */

#define TYPE_FORM                100
#define TYPE_TERM                101
#define TYPE_LISTCLAUSE          102 

/* Subtypes:      */

#define TYPE_MODFORM             200
#define TYPE_FOFORM              201
#define TYPE_CLAUSE              202


