

/********************************************************/
/* File inoutput.m                                      */
/* part of bliksem.                                     */
/********************************************************/


#define CLASS extern
#include "basics.h" 
#include "tokens.h"
#include "extfunc.h"
 


/**********************************************************************/
/* This is the scanner of the bliksem theorem prover.                 */
/* It is able to recognize the following tokens:                      */
/* 1. comments starting with / * and ending with * / (without space)  */
/* 2. variables consisting of letters, digits, and _?$%               */
/* 3. Tokens from the list tokenlist.                                 */
/* 4. EOF.                                                            */
/**********************************************************************/


/* The tokens are alfabetically sorted by function inittokens.    */

#define MAXTOKENS        50 
#define MAXTOKENSTRING   10
static struct 
{
   char string [ MAXTOKENSTRING ];
   int code; 
}
   tokenlist [ MAXTOKENS ];
static int freetokenlist; 


/* The entries of the tokens starting with a given letter.    */
/* This array is meaningful because the tokens are sorted.    */
/* If first and sup < 0, then there are no tokens starting    */
/* with the given letter.                                     */

static struct
{
   int first;
   int sup;
}
   tokenentries [ 128 ];


/* used to recognize the start of a comment: */
#define BLABLA  (-1)



/* add a token into the token list. */

static void maketoken ( s, nr )
char *s;
int nr; 
{
   (void) strcpy( tokenlist [ freetokenlist ]. string, s );
   tokenlist [ freetokenlist ]. code = nr;
   freetokenlist ++ ;
   if( freetokenlist >= MAXTOKENS - 3 )
      internalproblem( "increase MAXTOKENS" );
}

/* The token table. Tokens can be added here.   */ 

static void maketokens ( )
{
   freetokenlist = 0;

   maketoken( "<", TOK_LESS );
   maketoken( ">", TOK_GREATER );
   maketoken( "[", TOK_LSQBRACKET );
   maketoken( "]", TOK_RSQBRACKET );
   maketoken( "(", TOK_LPAR );
   maketoken( ")", TOK_RPAR );
   maketoken( "{", TOK_LBRACKET );
   maketoken( "}", TOK_RBRACKET );

   maketoken( "<>", TOK_POSSIBLE );
   maketoken( "[]", TOK_NECESSARY );
   
   maketoken( ",",   TOK_COMMA );
   maketoken( ".",   TOK_DOT );
   maketoken( "<->", TOK_EQUIV );
   maketoken( "->",  TOK_IMPLIES );
   maketoken( "|",   TOK_OR );
   maketoken( "&",   TOK_AND );
   maketoken( "!",   TOK_NOT );
   maketoken( "||",  TOK_FALSE );
   maketoken( "&&",  TOK_TRUE );

   maketoken( "=",    TOK_EQUAL );
   maketoken( "==>",  TOK_REWRITES );
   maketoken( "==",   TOK_EQUALEQUAL );
   maketoken( "!=",   TOK_NOTEQUAL );
   maketoken( "!==",  TOK_NOTEQUALEQUAL );
   maketoken( "<=",   TOK_LESSEQUAL );
   maketoken( ">=",   TOK_GREATEREQUAL );
   maketoken( ">>",   TOK_GREATERGREATER );
   maketoken( "<<",   TOK_LESSLESS );
   maketoken( ">>=",  TOK_GREATERGREATEREQUAL );
   maketoken( "<<=",  TOK_LESSLESSEQUAL );
   maketoken( ">>==", TOK_GREATERGREATEREQUALEQUAL );
   maketoken( "<<==", TOK_LESSLESSEQUALEQUAL );
   maketoken( ":",    TOK_IN );

   maketoken( "+", TOK_PLUS );
   maketoken( "-", TOK_MINUS );
   maketoken( "*", TOK_TIMES );
   maketoken( "/", TOK_DIVIDES );
   maketoken( "^", TOK_POW );

   maketoken( "{}", TOK_NIL ); 

   maketoken( "/*", BLABLA );
}


static void initoperators () 
{
   int i;

   arities [ OP_NIL ] = 0;
   arities [ OP_DOT ] = 2;
   
   arities [ OP_FALSE ] =    0;
   arities [ OP_TRUE ] =     0;
   arities [ OP_NOT ] =      1; 
   arities [ OP_AND ] =      2; 
   arities [ OP_OR ] =       2; 
   arities [ OP_IMPLIES ] =  2; 
   arities [ OP_EQUIV ] =    2; 
 
   arities [ OP_FORALL ] = 2;
   arities [ OP_EXISTS ] = 2;
   arities [ OP_NECESSARY ] = 1;
   arities [ OP_POSSIBLE ] = 1;
   
   arities [ OP_EQUAL ] = 2;
   arities [ OP_REWRITES ] = 2;
   
   arities [ OP_LESS ] = 2;
   arities [ OP_GREATER ] = 2;
   arities [ OP_EQUALEQUAL ] = 2; 
   arities [ OP_LESSEQUAL ] = 2;
   arities [ OP_GREATEREQUAL ] = 2;
   arities [ OP_GREATERGREATER ] = 2; 
   arities [ OP_LESSLESS ] = 2;
   arities [ OP_GREATERGREATEREQUAL ] = 2;
   arities [ OP_LESSLESSEQUAL ] =       2;
   arities [ OP_GREATERGREATEREQUALEQUAL ] = 2;
   arities [ OP_LESSLESSEQUALEQUAL ] = 2;
   arities [ OP_IN ] = 2;   /* In the end not used for sets. */ 

   arities [ OP_PLUS ] = 2;
   arities [ OP_MINUS ] = 2;
   arities [ OP_TIMES ] = 2;
   arities [ OP_DIVIDES ] = 2;
   arities [ OP_POW ] = 2;

   arities [ OP_UNARYMINUS ] = 1;
   arities [ OP_UNARYPLUS ] = 1;

   for( i = 0; i < OP_NONPRIMITIVE; i ++ ) 
   {
      operators [i]. flags = 0;
   }

   freeoperators = OP_NONPRIMITIVE;
}




/*******************************************************************/
/*          Functions for preparing the tokenlist.                 */
/*******************************************************************/

/* s1 < s2 if its first letter is smaller, or if the first letters */
/* are equal, and the length of s1 is less than the length of s2.  */

static int tokenless ( s1, s2 )
char *s1;
char *s2; 
{
   if( s1 [0] == 0 )
      exit(0);
   if( s2 [0] == 0 )
      exit(0);

   if( s1 [0] < s2 [0] )
      return 1;
   if( s1 [0] > s2 [0] )
      return 0;

   while( s1 [0] != 0 )
   {
      if( s2 [0] == 0 )
         return 0;
      s1 ++ ; s2 ++ ; 
   }
   if( s2 [0] != 0 ) 
      return 1;
   return 0;
}



static void inittokens ( )
{
   int i;
   char s [ MAXTOKENSTRING ];
   int c;

   maketokens ( ); 
   /* sort tokenlist. */ 
restart:
   for( i = 0; i < freetokenlist - 1; i ++ ) 
   {
      if( tokenless ( tokenlist [ i + 1 ]. string, 
                      tokenlist [ i ]. string ))
      {
         (void) strcpy( s, tokenlist [ i ]. string );
         (void) strcpy( tokenlist [ i ]. string,
                        tokenlist [ i + 1 ]. string );
         (void) strcpy( tokenlist [ i + 1 ]. string, s );             

         c = tokenlist [ i ]. code;
         tokenlist [ i ]. code = tokenlist [ i + 1 ]. code;
         tokenlist [ i + 1 ].code = c;
         goto restart;
      }
   }

   for( i = 0; i < 128; i ++ ) 
   {
      tokenentries [i]. first = -1;
      tokenentries [i]. sup = -1;
   }

   i = 0;
   while( i < freetokenlist )
   {
      c = tokenlist [i]. string [0];
      tokenentries [c]. first = i;
      i ++ ; 
      while( i < freetokenlist && 
             tokenlist [i]. string [0] ==
             tokenlist [ i - 1 ]. string [0] )
      {
         i ++ ; 
      }
      tokenentries [c]. sup = i; 
   }

#if 0
   for( i = 0; i < freetokenlist; i ++ ) 
   {
      printf("%d: %s    %d\n", tokenlist [i]. string [0], 
                        tokenlist [i]. string,
                        tokenlist [i]. code );
   }
   for( i = 0; i < 128; i ++ ) 
   {
      if( tokenentries [i]. first >= 0 )
         printf("[%d: %d, %d]\n", i, 
            tokenentries [i]. first,
            tokenentries [i]. sup );
   }
#endif
}
 


/********************************************************************/
/*         The scanner.                                             */
/********************************************************************/

static bliksemfile *scanfile;
extern int attribute; 

int syntaxerror( )
{
   errorwithline( "syntax error" );
   return 0;
}


static void readcharacter( f )
bliksemfile *f;
{
   if( f -> rw != FILE_READ )
   {
      errorwithoutline( "file not open for reading!" );
   }

   if( f -> lookahead == EOF )
      return ;

   ( f -> lookahead ) = getc ( f -> handle );
   if( f -> lookahead == '\n' )
   {
      ( f -> linenum ) ++ ;
      ( f -> linepos ) = 0;
   }
   else
      ( f -> linepos ) ++ ;
}
   

static int allowedinname ( x )
int x;
{
   if( x >= 'a' && x <= 'z' )
      return 1;
   if( x >= 'A' && x <= 'Z' )
      return 1;
   if( x >= '0' && x <= '9' )
      return 1;
   if( x == '_' || x == '?' || 
       x == '$' || x == '%' )
      return 1;
   return 0;
}

/* read an unquoted name: */

static void scanunquotedname ( f ) 
bliksemfile *f;
{
   char s [ MAXNAMELEN ];
   int i;

   i = 0;
   while( allowedinname ( f -> lookahead )) 
   {
      s[i] = ( f -> lookahead );
      i ++ ; 
      if( i >= MAXNAMELEN - 3 )
         errorwithline( "name is too long" );
      readcharacter( f ); 
   }
   s[i] = 0;
   attribute = makename ( s ); 
}


/* read a quoted name. (")     */

static void scanquotedname( f )
bliksemfile *f;
{
   char s [ MAXNAMELEN ];
   int i;

   readcharacter(f);

   i = 0;
   while( (f -> lookahead) != EOF && 
          (f -> lookahead) != '\n'  &&
          (f -> lookahead) != '\"' )
   {
      s[i] = ( f -> lookahead );
      i ++ ; 
      if( i >= MAXNAMELEN - 3 )
         errorwithline( "name is too long" );
      readcharacter(f);
   }
   if( ( f -> lookahead ) != '\"' )
      errorwithline( "newline or EOF in string" ); 
   readcharacter(f); 
   s[i] = 0;
   attribute = makename( s );
}


/* return 1 if s1 is an initial string of s2. */ 

static int initialstring( s1, s2 )
char *s1;
char *s2;
{
   while( *s1 != 0 )
   {
      if( *s1 != *s2 )
         return 0;
      s1 ++ ; s2 ++ ; 
   }
   return 1;
}

/* try to scan a special token from tokenlist [].  */ 

static int scanspecialtoken( f )
bliksemfile *f;
{
   char s[ MAXTOKENSTRING + 2 ];

   int i; 
   int start, end;
   int length;

   start = tokenentries [ f -> lookahead ]. first; 
   end = tokenentries [ f -> lookahead ]. sup;

   s[0] = (f -> lookahead);
   s[1] = 0;
   length = 1;

next:
   readcharacter(f);
   s [ length ++ ] = ( f -> lookahead );
   s [ length ] = 0;

#if 0 
   printf("trying [%s] ", s ); 
#endif
   for( i = start; i < end; i ++ ) 
   {
      if( initialstring( s, tokenlist [i]. string ))
      {
         start = i;
         goto next;
      } 
   }
   
   length -- ;
   s [ length ] = 0;
#if 0
   printf("we decide [%s]\n", s );
#endif    
   for( i = start; i < end; i ++ ) 
   {
      if( strcmp( s, tokenlist [i]. string ) == 0 )
         return tokenlist [i]. code;
   }

   errorwithline( "scanner error" );
   return 0; 
}


/* scan a comment. The initial / * has already been read. */ 

static void scancomment ( f )
bliksemfile *f;
{
   while( 1 )
   {
      while( (f -> lookahead) != EOF && 
             (f -> lookahead) != '*' )
      {
         readcharacter(f); 
      }
      if( (f -> lookahead) == EOF ) 
         errorwithline( "end of file in comment" );
      readcharacter(f);

      if( (f -> lookahead) == '/' )
      {
         readcharacter(f); 
         return;
      }
      if( (f -> lookahead) == EOF )
         errorwithline( "end of file in comment" );
   }
}


/***********************************************************/
/*        read a token from the input.                     */ 
/***********************************************************/


int scanner ( ) 
{
   int a;

aftercomment: 
   if( scanfile -> lookahead == ' ' ||
       scanfile -> lookahead == '\n' || 
       scanfile -> lookahead == '\t' )
   {
      readcharacter( scanfile );
      goto aftercomment; 
   }

   /* This type of comment has been added for the tptp2X utility. */
 
   if( scanfile -> lookahead == '#' )
   {
      while( scanfile -> lookahead != '\n' &&
             scanfile -> lookahead != EOF ) 
      {
         readcharacter( scanfile );
      }
      if( scanfile -> lookahead == EOF ) 
      {
         errorwithline( "end of file in comment" );
      }
      readcharacter( scanfile );
      goto aftercomment; 
   }
 
   attribute = 0;  

   if( (scanfile -> lookahead) == EOF )
      return TOK_EOF;

   if( (scanfile -> lookahead) >= 0X80 || 
       (scanfile -> lookahead) < 0 )
      errorwithline( "non-ascii character in input" );

   if( allowedinname( scanfile -> lookahead ))
   {
      scanunquotedname ( scanfile );
      return TOK_NAME;
   }
 
   if( (scanfile -> lookahead ) == '\"' )
   {
      scanquotedname( scanfile );
      return TOK_NAME;
   }

   if( tokenentries [ scanfile -> lookahead ]. first >= 0 )
   {
      a = scanspecialtoken( scanfile );
      if( a == BLABLA )
      {
         scancomment( scanfile );
         goto aftercomment; 
      }
      return a; 
   }
   errorwithline( "garbage in input" ); 
}
 

/**************************************************************/
/*           Maphoon declarations:                            */
/**************************************************************/ 


%noemparser parser
%noemscanner scanner
%noemerror syntaxerror
%noemattr attribute
%noemstype int
%geentekendefs
%redvooruit

%begintaal

%eindsym TOK_NAME ; 

%eindsym TOK_LESS TOK_GREATER TOK_LSQBRACKET TOK_RSQBRACKET ;
%eindsym TOK_LPAR TOK_RPAR TOK_LBRACKET TOK_RBRACKET ;

%eindsym TOK_POSSIBLE TOK_NECESSARY ;

%eindsym TOK_COMMA TOK_DOT TOK_EQUIV TOK_IMPLIES ;
%eindsym TOK_OR TOK_AND TOK_NOT TOK_FALSE TOK_TRUE ;

%eindsym TOK_EQUAL TOK_REWRITES TOK_EQUALEQUAL TOK_NOTEQUAL ;
%eindsym TOK_NOTEQUALEQUAL TOK_LESSEQUAL TOK_GREATEREQUAL ;
%eindsym TOK_GREATERGREATER TOK_LESSLESS ; 
%eindsym TOK_GREATERGREATEREQUAL TOK_LESSLESSEQUAL ;
%eindsym TOK_GREATERGREATEREQUALEQUAL TOK_LESSLESSEQUALEQUAL ; 
%eindsym TOK_IN ; 

%eindsym TOK_PLUS TOK_MINUS TOK_TIMES TOK_DIVIDES TOK_POW ;

%eindsym TOK_NIL ;
%eindsym TOK_NOSPACE TOK_EOF ; 

%startsym Input; 

   int n; 
   int op;

% Input    : Commands TOK_EOF 
   return 0;
%          ;

% Commands : Command
%          | Commands Command
%          ;

% Command     : Expression TOK_DOT
%             ;

%nietassoc

% Expression     : TOK_MINUS Expression

   $$ = $2;
   insertoperator( OP_UNARYMINUS, $2 );

%                | TOK_PLUS Expression

   $$ = $2;
   insertoperator( OP_UNARYPLUS, $2 );

%rechtsassoc 
%                | Expression TOK_POW Expression

   $$ = $1;
   insertoperator( OP_POW, $1 );

%linksassoc
%                | Expression TOK_TIMES Expression

   $$ = $1;
   insertoperator( OP_TIMES, $1 );

%                | Expression TOK_DIVIDES Expression

   $$ = $1;
   insertoperator( OP_DIVIDES, $1 );

%linksassoc
%                | Expression TOK_PLUS Expression

   $$ = $1;
   insertoperator( OP_PLUS, $1 );

%                | Expression TOK_MINUS Expression

   $$ = $1;
   insertoperator( OP_MINUS, $1 );

%nietassoc
%                | Expression TOK_EQUAL Expression

   $$ = $1;
   insertoperator( OP_EQUAL, $1 ); 

%                | Expression TOK_REWRITES Expression

   $$ = $1;
   insertoperator( OP_REWRITES, $1 );

%                | Expression TOK_LESS Expression

   $$ = $1;
   insertoperator( OP_LESS, $1 );

%                | Expression TOK_GREATER Expression

   $$ = $1;
   insertoperator( OP_GREATER, $1 );

%                | Expression TOK_EQUALEQUAL Expression

   $$ = $1; 
   insertoperator( OP_EQUALEQUAL, $1 );

%                | Expression TOK_NOTEQUAL Expression

   $$ = $1; 
   insertoperator( OP_EQUAL, $1 );
   insertoperator( OP_NOT, $1 );

%                | Expression TOK_NOTEQUALEQUAL Expression

   $$ = $1;
   insertoperator( OP_EQUALEQUAL, $1 );
   insertoperator( OP_NOT, $1 );

%                | Expression TOK_LESSEQUAL Expression

   $$ = $1;
   insertoperator( OP_LESSEQUAL, $1 );

%                | Expression TOK_GREATEREQUAL Expression

   $$ = $1;
   insertoperator( OP_GREATEREQUAL, $1 );

%                | Expression TOK_GREATERGREATER Expression

   $$ = $1;
   insertoperator( OP_GREATERGREATER, $1 );

%                | Expression TOK_LESSLESS Expression

   $$ = $1;
   insertoperator( OP_LESSLESS, $1 );

%                | Expression TOK_GREATERGREATEREQUAL Expression

   $$ = $1;
   insertoperator( OP_GREATERGREATEREQUAL, $1 );

%                | Expression TOK_LESSLESSEQUAL Expression
 
   $$ = $1;
   insertoperator( OP_LESSLESSEQUAL, $1 );

%                | Expression TOK_GREATERGREATEREQUALEQUAL Expression

   $$ = $1;
   insertoperator( OP_GREATERGREATEREQUALEQUAL, $1 ); 

%                | Expression TOK_LESSLESSEQUALEQUAL Expression

   $$ = $1;
   insertoperator( OP_LESSLESSEQUALEQUAL, $1 );

%                | Expression TOK_IN Expression

   $$ = $1;
   insertoperator( OP_IN, $1 );

%nietassoc
%                | TOK_NOT Expression

   $$ = $2;
   insertoperator( OP_NOT, $2 );    

%linksassoc
%                | Expression TOK_AND Expression

   $$ = $1;
   insertoperator( OP_AND, $1 ); 

%linksassoc   
%                | Expression TOK_OR Expression

   $$ = $1;
   insertoperator( OP_OR, $1 );
   
%linksassoc 
%                | Expression TOK_EQUIV Expression

   $$ = $1;
   insertoperator( OP_EQUIV, $1 );

%rechtsassoc 
%                | Expression TOK_IMPLIES Expression

   $$ = $1;
   insertoperator( OP_IMPLIES, $1 ); 

%nietassoc
%                | TOK_POSSIBLE Expression

   $$ = $2;
   insertoperator( OP_POSSIBLE, $2 );

%                | TOK_NECESSARY Expression

   $$ = $2;
   insertoperator( OP_NECESSARY, $2 );

%                | TOK_LSQBRACKET Universallist TOK_RSQBRACKET Expression 

   $$ = $2;

%                | TOK_LESS Existentiallist TOK_GREATER Expression

   $$ = $2;

%geenoperator

%                | TOK_LPAR Expression TOK_RPAR

   $$ = $2; 

%                | TOK_NIL

   $$ = freetermspace;
   insertoperator( OP_NIL, freetermspace );

%                | TOK_LBRACKET List TOK_RBRACKET

    $$ = $2;

%                | TOK_NAME 

   $$ = freetermspace; 
   op = makeuseroperator( $1, 0 );
   insertoperator( op, freetermspace ); 

%                | TOK_NAME TOK_LPAR Arglist TOK_RPAR 

   n = countterms( $3 );
   op = makeuseroperator( $1, n );
   $$ = $3;
   insertoperator( op, $3 );

%                | TOK_TRUE

   $$ = freetermspace;
   insertoperator( OP_TRUE, freetermspace );

%                | TOK_FALSE

   $$ = freetermspace;
   insertoperator( OP_FALSE, freetermspace ); 

%                ; 


% Universallist    :   TOK_NAME
  
   $$ = freetermspace;  
   op = makeuseroperator( $1, 0 );
   insertoperator( op, freetermspace ); 
   insertoperator( OP_FORALL, $$ );

%                  |   TOK_NAME TOK_COMMA Universallist

   op = makeuseroperator( $1, 0 );
   $$ = $3;
   insertoperator( op, $3 ); 
   insertoperator( OP_FORALL, $3 ); 

%                  ;

% Existentiallist : TOK_NAME

   $$ = freetermspace;
   op = makeuseroperator( $1, 0 );
   insertoperator( op, freetermspace );
   insertoperator( OP_EXISTS, $$ );

%                 | TOK_NAME TOK_COMMA Existentiallist 

   op = makeuseroperator( $1, 0 );
   $$ = $3;
   insertoperator( op, $3 );
   insertoperator( OP_EXISTS, $3 );

%                 ;

% List            : Expression

   $$ = $1;
   insertoperator( OP_NIL, freetermspace );
   insertoperator( OP_DOT, $$ ); 

%                 | Expression TOK_COMMA List

   $$ = $1;
   insertoperator( OP_DOT, $1 );

%                 ;

% Arglist    : Expression
   $$ = $1; 
%            | Arglist TOK_COMMA Expression
   $$ = $1; 
%            ;


%eindetaal


 

/******************************************************************/
/*               the output functions:                            */
/******************************************************************/



/* only to be called from printstring.    */

static void printchar ( f, x )
bliksemfile *f;
int x;
{
   int c;
   char s [ MAXFILENAMELEN + 50 ];

   if( (f -> rw) != FILE_WRITE )
      errorwithoutline( "printstring: file not open for writing" );
   c = putc( x, f -> handle );
   if( c == EOF )
   {
      sprintf( s, "problems with writing into file %s", f -> name );
      errorwithoutline( s );
   } 
   if( x == '\n' )
   {
      (f -> linenum) ++ ; 
      (f -> linepos) = 0;
   }
   else
      (f -> linepos) ++ ; 
}


/* print a string into a file: */  

void printstring( f, s )
bliksemfile *f;
char *s;
{
   int length;

   length = 0;
   while( s [ length ] )
      length ++ ;

   if( (f -> linepos) + length > MAXLINEPOS )
   {
      printchar( f, '\n' ); 
      printchar( f, ' ' );
      printchar( f, ' ' );
   }        
   while( *s )
   {
      printchar( f, *s );
      s ++ ; 
   }
}


/* print a string into a file. The difference between       */
/* printstring( f, s ), and printstringnnl( f, s ) is       */
/* that printstring prints a newline when the string is     */
/* going to be on a position >= MAXLINEPOS. When printing   */
/* prolog-output there can be no whitespace before a (.     */


void printstringnnl( f, s )
bliksemfile *f;
char *s;
{
   while( *s )
   {
      printchar( f, *s );
      s ++ ; 
   }
}


 
/* print a variable: */ 

static char nicevars [] = "XYZTUW"; 
 
void printvariable( f, x )
bliksemfile *f; 
int x;
{
   char printedvar [20];

   if( x >= 0 ) 
   {
      internalproblem( "Not a variable" );
   }

   x = -x;
   x --; 

   /* Now we have a positive integer 0 <= x <= ....         */

   switch( prologoutput )
   {
   case 0:
   case 1:
      if( x < 6 )
      {
         printedvar [0] = nicevars [x];
         printedvar [1] = 0;
      } 
      else
         sprintf( printedvar, "V%d", x - 6 );
      break;
   case 2:
      sprintf( printedvar, "vvvv_(%d)", x );
      break;  
   case 3:
      sprintf( printedvar, "vvvv_%d", x ); 
      break; 
   default:
      internalproblem( "incorrect value of prologoutput" );
   } 
   printstring( f, printedvar );
}


/* We try to be on the safe side.   */

static int allowedinprologname( x )
int x;
{
   if( x >= 'a' && x <= 'z' )
      return 1;
   if( x >= 'A' && x <= 'Z' )
      return 1;
   if( x >= '0' && x <= '9' )
      return 1;
   return 0;
}


/* print a name:    */

void printname ( f, n )
bliksemfile *f; 
int n;
{
   char *s;
   int i;
   int special;
   char squoted [ MAXNAMELEN + 5 ];
   int quote;

   s = namespace + nametable [ n ]. string; 

   /* Decide whether or not quotes are necessary, */
   /* and what the quote will be.                 */

   if( prologoutput )
   {
      special = 0; 
      for( i = 0; s[i]; i ++ )
      {
         if( ! allowedinprologname( s[i] ))
            special = 1; 
      }
      if( ! ( s[0] >= 'a' && s[0] <= 'z' ))
         special = 1;
      quote = '\'';
   }
   else
   {
      special = 0; 
      for( i = 0; s[i]; i ++ ) 
      {
         if( ! allowedinname( s[i] ))
            special = 1;
      }
      quote = '\"';
   }
   if( special )
   {
      for( i = 0; s[i]; i ++ )
      {
         squoted [ i + 1 ] = s[i];
      }
      squoted[0] = quote;
      squoted[i+1] = quote;
      squoted[i+2] = 0;
      printstring( f, squoted );
   }
   else
      printstring( f, s ); 
}


/* print an integer:                */ 

void printint( f, x )
bliksemfile *f;
int x;
{
   char s [ 40 ];

   sprintf( s, "%d", x );
   printstring( f, s );
}


/* print an integer hexadecimal.    */

void printhex( f, x )
bliksemfile *f;
int x;
{
   int k;
   int i;
   char s[10];
   int m;
  
   i = 0;
   for( k = sizeof(int) * 2 - 1; k >= 0; k -- )
   {
      if( k == sizeof(int) - 1 )
         s [ i ++ ] = '.';

      m = x >> ( k * 4 );
      m = m & 0X0F;
      if( m >= 10 )
      {
         m -= 10;
         s [ i ++ ] = 'A' + m;
      }
      else
      {
         s [ i ++ ] = '0' + m;
      }
   }
   s[i] = '\0';
   printstring( f, s );
}
 

void printnametable ( f )
bliksemfile *f; 
{
   int i, k;

   printstring( f, "The name table:\n\n" );
   for( i = 0; i < freenametable; i ++ ) 
   {
      printname( f, i );
      printstring( f, "  " ); 
   }
   printstring( f, "\n\n" ); 

#if 1 
   printf("Name table:/n/n");   
   for( i = 0; i < MAXHASHVALUE; i ++ ) 
   {
      if( hashindex [i] >= 0 )
      {
         printf("hashvalue %d:\n", i );
         k = hashindex [i];
         while( k >= 0 )
         {
            printf("   %d %s\n", k, namespace + nametable [k]. string );
            k = nametable [k]. next;
         }
      }
   }
#endif 
}



/* strings used in printing the operators: */ 

static char *operatorstrings [] = 
{
   "{}", ".", "||", "&&",
   "!", "&", "|", "->",
   "<->",

   "[]", "<>", "[]", "<>",
   "=", "==>", 

   "<", ">", "==", "<=",
   ">=", ">>", "<<",
   ">>=", "<<=", ">>==", "<<==",
   ":",

   "+", "-", "*", "/", 
   "^", "-", "+"

};


#define LEFTASSOC       (0X400)  
#define RIGHTASSOC      (0X800)
#define NOTASSOC        (0XC00)

/* an operator is leftassoc if it is able to defend its arguments   */
/* against right context of the same binding strength.              */
/* An operator is rightassoc if it is able to defend its arguments  */
/* against left context of the same binding strength.               */

static int operatorpriorities [] = 
{
   0, 0, 0, 0,         /* irrelevant. */ 
   
   80, 70 + LEFTASSOC, 60 + LEFTASSOC, 10 + RIGHTASSOC,  
   50 + LEFTASSOC, 
   10, 10, 10, 10,
   100 + NOTASSOC, 100 + NOTASSOC, 
 
   100 + NOTASSOC, 100 + NOTASSOC, 100 + NOTASSOC, 100 + NOTASSOC,  
   100 + NOTASSOC, 100 + NOTASSOC, 100 + NOTASSOC,
   100 + NOTASSOC, 100 + NOTASSOC, 100 + NOTASSOC, 100 + NOTASSOC, 
   100 + NOTASSOC,

   110 + LEFTASSOC, 110 + LEFTASSOC, 120 + LEFTASSOC, 120 + LEFTASSOC, 
   130 + RIGHTASSOC, 140, 140

};



/* printtermcontext( f, term, left, right ) prints term into file f. */
/* It uses left and right to decide if it is necessary to            */
/* insert parentheses.                                               */


static void printtermcontext( f, term, left, right )
bliksemfile *f;
int term;
int left;
int right;
{
   int operator;
   int arity;
   int i;
   int pars;
   int priority;

   operator = termspace [ term ];
   if( operator < 0 ) 
   {
      printvariable ( f, operator );
      return;
   }

   if( operator >= freeoperators && operator >= OP_NONPRIMITIVE )
      goto error;

   if( operator == OP_DOT ) 
   {
      /* print a list. */

      printstring( f, "{ " );
      while( operator == OP_DOT )
      { 
         printtermcontext( f, term + 1, 0, 0 ); 
         term = termends [ term + 1 ];
         operator = termspace [ term ]; 
         if( operator == OP_DOT )
            printstring( f, ", " );
         else
            printstring( f, " }");
      }
      if( operator != OP_NIL ) 
         goto error; 
      return;  
   } 

   if( arities [ operator ] == 0 )
   {
      if( operator >= OP_NONPRIMITIVE ) 
         printname( f, operators [ operator ]. name );
      else
         printstring( f, operatorstrings [ operator ] );
      return;
   }
   
   if( operator >= OP_NONPRIMITIVE )
   {
      /* print a simple functional term. */
      /* We can be sure that the arity is greater than 0. */

      printname( f, operators [ operator ]. name );
      printstring( f, "( " );
      arity = arities [ operator ]; 
      term ++ ; 
      for( i = 0; i < arity; i ++ ) 
      {
         printtermcontext( f, term, 0, 0 );
         if( i + 1 < arity )
            printstring( f,", " );
         else
            printstring( f, " )" );
         term = termends [ term ];
      }
      return;
   }
   /* what remain are the predefined, non-zero arity operators. */

   if( operator == OP_FORALL || operator == OP_EXISTS ) 
   {
      /* the quantifiers: */

      pars = 0;
      priority = operatorpriorities [ operator ] & 0X3FF; 
      if( right >= priority )
         pars = 1;

      if( pars )
         printstring( f, "( "); 

      switch( operator )
      {
      case OP_FORALL: 
         printstring( f, "[ " );
         break;
      case OP_EXISTS:
         printstring( f, "< " );    
         break;
      }
      printtermcontext( f, term + 1, 0, 0 );
      term = termends [ term + 1 ];
      while( termspace [ term ] == operator )
      {  
         /* try to combine equal quantifiers as in [ .., .., .. ] */ 
         printstring( f, ", " );
         printtermcontext( f, term + 1, 0, 0 );
         term = termends [ term + 1 ];
      }
     
      switch( operator )
      {
      case OP_FORALL:  
         printstring( f, " ] " );
         break;
      case OP_EXISTS:
         printstring( f, " > " );
         break;
      }
 
      if( pars )
         printtermcontext( f, term, priority, 0 );
      else
         printtermcontext( f, term, priority, right );
   
      if( pars )
         printstring( f, " )");
       
      return; 
   }

   if( arities [ operator ] == 1 )
   {
      /* unary operators: */
      
      pars = 0;
      priority = operatorpriorities [ operator ] & 0X3FF; 

      if( right >= priority )
         pars = 1;

      if( pars )
         printstring( f, "( ");

      printstring( f, operatorstrings [ operator ] );
      printstring( f, " " );
   
      term ++ ; 
      if( pars )
         printtermcontext( f, term, priority, 0 );
      else
         printtermcontext( f, term, priority, right );
      
      if( pars )
         printstring( f, " )"); 
      return; 
   }

   if( arities [ operator ] != 2 )
      goto error; 

   pars = 0;
   priority = operatorpriorities [ operator ] & 0X3FF;

   switch( operatorpriorities [ operator ] & 0XFC00 )
   {
   case LEFTASSOC:
      if( left >= priority ) 
         pars = 1; 
      if( right > priority )
         pars = 1;
      break; 
   case RIGHTASSOC:
      if( left > priority )
         pars = 1; 
      if( right >= priority ) 
         pars = 1;  
      break; 
   case NOTASSOC:
      if( left >= priority )
         pars = 1;
      if( right >= priority ) 
         pars = 1; 
      break; 
   default:
      goto error; 
   }

   if( pars )
   {
      printstring( f, "( ");
      printtermcontext( f, term + 1, 0, priority );
      printstring( f, " " );
      printstring( f, operatorstrings [ operator ] );
      printstring( f, " " );
      printtermcontext( f, termends [ term + 1 ], priority, 0 );
      printstring( f, " )" );
   }
   else
   {
      printtermcontext( f, term + 1, left, priority );
      printstring( f, " " );
      printstring( f, operatorstrings [ operator ] );
      printstring( f, " " );
      printtermcontext( f, termends [ term + 1 ], priority, right );
   }
   return; 

error:
   internalproblem( "there is a deep, deep problem" );
}


/* print a term in prolog readable format.             */
/* This is used when the flag prologoutput > 0.        */
/* We try to use TPTP-format.                          */


static void printprologterm( f, term )
bliksemfile *f;
int term;
{
   int operator;
   int i, ar;

   operator = termspace [ term ];
   ar = arities [ operator ];

   /* Check for variable.           */

   if( operator < 0 )
   {
      printvariable( f, operator );
      return;
   }

   /* Check for list:                  */

   if( operator == OP_NIL )
   {
      printstring( f, "[]" );
      return;
   }

   if( operator == OP_DOT )
   {
      printstring( f, "[ " );
      while( operator == OP_DOT )
      {
         printprologterm( f, term + 1 );
         term = termends [ term + 1 ];
         operator = termspace [ term ];
         if( operator == OP_DOT )
            printstring( f, ", " );
         else
            printstring( f, " ]" );
      }
      if( operator != OP_NIL )
         internalproblem( "nil expected" );
      return;
   }

   /* Check for first order operators. */

   switch( operator )
   {
   case OP_FALSE:
      printstring( f, "false" );
      break; 
   case OP_TRUE:
      printstring( f, "true" );
      break;
   case OP_NOT:
      printstring( f, "~( " );
      printprologterm( f, term + 1 );
      printstring( f, " )");
      return;
   case OP_AND:
   case OP_OR:
   case OP_IMPLIES:
   case OP_EQUIV:
      if( operator == OP_AND ) printstring( f, "&( " );
      if( operator == OP_OR ) printstring( f, "|( " );
      if( operator == OP_IMPLIES ) printstring( f, "=>( " );
      if( operator == OP_EQUIV ) printstring( f, "<=>( " );

      printprologterm( f, term + 1 );
      printstring( f, ", " );
      printprologterm( f, termends [ term + 1 ] );
      printstring( f, " )" ); 
      return;
   case OP_FORALL:
   case OP_EXISTS:
      printstring( f, ":( ");
      if( operator == OP_FORALL ) printstring( f, "!( " );
      if( operator == OP_EXISTS ) printstring( f, "?( " );
      printstring( f, "[ ");
      printprologterm( f, term + 1 );
      printstring( f, " ] ), " );
      printprologterm( f, termends [ term + 1 ] );
      printstring( f, " )");
      return;
   case OP_REWRITES:
   case OP_EQUAL:
      printstring( f, "=( ");
      printprologterm( f, term + 1 );
      printstring( f, ", " );
      printprologterm( f, termends [ term + 1 ] );
      printstring( f, " )" );
      return; 
   }

   /* The default case. Print the operator.   */

   if( operator >= OP_NONPRIMITIVE )
      printname( f, operators [ operator ]. name );
   else
      printstring( f, operatorstrings [ operator ] );
   
   if( ar > 0 )
   {
      printstringnnl( f, "( " );   /* It was prolog. Remember? */

      term = term + 1;
      for( i = 0; i < ar; i ++ )
      {
         printprologterm( f, term );
         if( i < ar - 1 )
            printstring( f, ", " );
         else
            printstring( f, " )" );
         term = termends [ term ];
      } 
   }
}


/* print a term. */ 

void printterm( f, term )
bliksemfile *f;
int term;
{
   if( termends [ term ] - term > 5000 )
   {
      printstring( f, "<TERM WITH SIZE " ); 
      printint( f, termends [ term ] - term );
      printstring( f, " OMITTED>" ); 
      return;
   }

   if( prologoutput )
      printprologterm( f, term );
   else
      printtermcontext( f, term, 0, 0 ); 
}





 
/* print a consecutive list of terms:                  */

void printtermlist( f, s, termstart, termend )
bliksemfile *f;
char *s;
int termstart;
int termend; 
{
   int i;

   if( prologoutput )
   {
      printstring( f, s ); 
      printstring( f, "\n" );

      if( termstart == termend )
         printstring( f, "[]" );
      else
      {
         printstring( f, "[\n   " );
         for( i = termstart; i < termend; i = termends [i] )
         {
            if( i != termstart )
               printstring( f, ",\n   " );
            printterm( f, i );
         }
         printstring( f, "\n]");
      }
      printstring( f, " .\n\n" );
   }
   else
   {
      printstring( f, "\n" ); 
      printstring( f, s );
      printstring( f, "\n\n" );

      for( i = termstart; i < termend; i = termends [i] )
      {
         printterm( f, i );
         printstring( f, ".\n" );
      }
   }
}


/* print an operator:    */ 

void printoperator( f, op )
bliksemfile *f;
int op;
{
   if( op < OP_NONPRIMITIVE ) 
      printstring( f, operatorstrings [ op ] );
   else
      printname( f, operators [ op ]. name );
}


#if 0
/* Print the list of operators:   */

void printoperators( f )
bliksemfile *f;
{
   int i;

   printstring( f, "\n" );
   printstring( f, "The names in the input: \n\n");
  
   for( i = OP_NONPRIMITIVE; i < freeoperators; i ++ ) 
   {
      printoperator( f, i );
      printstring( f, ": " );
      printint( f, arities [i] );
      printstring( f, ",    " ); 
   }
   printstring( f, "\n\n");
}
#endif

/*********************************************************************/
/*     initialize the file structures.                               */
/*********************************************************************/
 

void initinoutput ( position, argc, argv )
int position;
int argc;
char *argv [];
{
   int i;
   
   if( position < argc - 1 )
   {
      printf("\ntoo many arguments:\n\n");
      for( i = 1; i < argc; i ++ )
         printf("%s\n", argv [i] );
      printf("\n");
      exit(1); 
   }
  
   if( position == argc )
   { 
      inputfile. handle = stdin; 
      strcpy( inputfile. name, "STDIN" );
   }
   else
   {
      if( strlen( argv [ position ] ) > MAXFILENAMELEN - 4 ) 
      {
         printf( "\nfilename %s is too long!\n\n", 
                 argv [ position ] );
         exit(1);
      }

      inputfile. handle = fopen( argv [ position ], "r" );
      if( inputfile. handle == ( FILE* ) NULL ) 
      {
         printf( "\ncould not open inputfile %s\n\n", 
                   argv [ position ] );
         exit(1);
      }   
      strcpy( inputfile. name, argv [ position ] );
   }
   inputfile. rw = FILE_READ;
   inputfile. linenum = 1;
   inputfile. linepos = 0;
   inputfile. lookahead = ' ';

   outputfile. handle = stdout;
   strcpy( outputfile. name, "STDOUT" );
   outputfile. rw = FILE_WRITE;
   outputfile. linenum = 1;
   outputfile. linepos = 0;
   outputfile. lookahead = ' ';

   for( i = 0; i < MAXHASHVALUE; i ++ ) 
      hashindex [i] = -1;
 
   freenamespace = 0;
   freenametable = 0;
 
   inittokens( );
   initoperators( );

   scanfile = &inputfile; 
}
 

