

/*****************************************************************/
/* file subsumption.c contains the functions that are related to */
/* subsumption.                                                  */
/*****************************************************************/


#define CLASS extern
#include "basics.h"
#include "resolution.h"
#include "extfunc.h"


#define MAXASSIGNMENTS 2000
static struct
{
   int var; 
   int term;
}
   assignments [ MAXASSIGNMENTS ];
static int freeassignments;

/* A matching is a set of simple assignments.                    */
/* matchinglists is an array containing lists of matchings.      */
/* assignments [ matchingstart ] ... assignments [ matchingend ] */
/* contains a matching. matchinglists [ rest ]                   */
/* refers to the cdr of the list, -1 is used for nil.            */
/* The list gives the possible alternatives for matching         */
/* a certain literal.                                            */


#define MAXMATCHINGLISTS 500
static struct
{
   int matchingstart;   /* references into assignments.             */ 
   int matchingend;   
   int rest;            /* reference into matchinglists, -1 on end. */ 
}
   matchinglists [ MAXMATCHINGLISTS ];
static int freematchinglists;


static struct
{
   int matchinglist; /* reference into matchinglists, -1 if there is    */ 
                     /* none.                                           */
   int firstfit;     /* clauses [ firstfit ] is the first literal       */
                     /* that matches.                                   */
   int firstdir;     /* the direction in which the first literal        */
                     /* matches, only used when substype = SUBS_EQREWR. */
   int supdir;       /* The maximal value that firstdir can take. It    */
                     /* is 2 for equality/rewrite literals, 1 for       */
                     /* other literals. Used only when substype =       */
                     /* SUBS_EQREWR.                                    */
   int nralternatives;
   int nrvariables;  /* nr of variables in the matched literal.         */
   int notinsemtab;
}
   litmatchings [ MAXCLAUSELEN ];
static int freelitmatchings;
   /* equal to the length of the first clause. */


static struct
{
   int conflict;    /* used in conflict. */ 
   int m;           /* the matching.     */ 
}
   semtab [ MAXCLAUSELEN ];

   /* We use a semantic tableaux for the subsumption algorithm. */


/*****************************************************************/
/* Matching for flatterms:                                       */
/*****************************************************************/


/* return 1 if term1 and term2 are equal. */ 


static int termsareequal( term1, term2 )
int term1;
int term2;
{
   int endterm1;

   int a; 
#if 0 
   printstring( &outputfile, "termsareequal " );
   printterm( &outputfile, term1 );
   printstring( &outputfile, " = " ); 
   printterm( &outputfile, term2 );
   printstring( &outputfile, "?\n" ); 
#endif
  
   endterm1 = termends [ term1 ]; 
   while( term1 < endterm1 )
   {
      a = termspace [ term1 ++ ];
      if( a != termspace [ term2 ++ ] )
         return 0;
   }
   return 1;
}



#if 0

static void printmatching( f, m )
bliksemfile *f;
int m;
{
   int i;

   for( i = matchinglists [m]. matchingstart;
        i < matchinglists [m]. matchingend;
        i ++ ) 
   {
      printvariable( f, assignments [i]. var );
      printstring( f, ":=" );
      printterm( f, assignments [i]. term );  
      if( i + 1 != matchinglists [m]. matchingend )
      {
         printstring( f, ", "); 
      }
   }
}


static void printlitmatchings ( f, clause2start )
bliksemfile *f;
int clause2start;
{
   int i, j, k;

   printstring( f, "Matching lists:\n" );

   for( i = 0; i < freelitmatchings; i ++ ) 
   {
      printstring( f, "   (F" );
      printint( f, litmatchings [i]. firstfit - clause2start );
      printstring( f, ", D" );
      printint( f, litmatchings [i]. firstdir ); 
      printstring( f, ", E" );
      printint( f, litmatchings [i]. supdir ); 
      printstring( f, ", A" );
      printint( f, litmatchings [i]. nralternatives );
      printstring( f, ", V" );
      printint( f, litmatchings [i]. nrvariables ); 
      printstring( f, ", S" );
      printint( f, litmatchings [i]. notinsemtab );
      printstring( f, ")   " );

      k = litmatchings [i]. matchinglist;
      while( k != -1 )
      {
         printmatching( f, k );

         if( matchinglists [k]. rest != -1 )
            printstring( f, " | " );
         k = matchinglists [k]. rest; 
      }
      printstring( f, "\n" );
   }
}

#endif 



/* Check if the matching in matchinglists[m] is internally consistent, */
/* that means does not assign different values to the same variable.   */
/* It also removes removes double, consistent assignments to           */
/* variables.                                                          */


static int matchingconsistent( m )
int m;
{
   int b, e;
   int i1, i2;
   int j;

   b = matchinglists [m]. matchingstart;
   e = matchinglists [m]. matchingend;
   
   i2 = b;
   for( i1 = b; i1 < e; i1 ++ )
   {
      for( j = b; j < i1; j ++ )
      {
         if( assignments [i1]. var == assignments [j]. var )
         {
            if( ! termsareequal( assignments [i1]. term,
                                 assignments [j]. term ))
            {
               return 0;
            }
            goto next;
         }
      }
      assignments [ i2 ]. var = assignments [i1].var;
      assignments [ i2 ++ ]. term = assignments [i1]. term;
   next:
      ;
   }    
   matchinglists [m]. matchingend = i2; 
   return 1;
}



/* Check if maching [m] is a renaming. This works only */
/* well if matching [m] has been normalized by         */
/* matchingconsistent.                                 */ 


static int matchingrenaming( m )
int m;
{
   int b, e;
   int i, j;

   b = matchinglists [m]. matchingstart;
   e = matchinglists [m]. matchingend;

   for( i = b; i < e; i ++ )
   {
      for( j = b; j < i; j ++ )
      {
         if( termspace [ assignments [i]. term ] == 
             termspace [ assignments [j]. term ] )
         {
            return 0;
         }
      }
   }
   return 1;
}


/* return 1 if matchlists m1 and m2 are consistent.      */
/* both m1 and m2 are references into matchlists.        */ 
/* matchlists are consistent if they do not assign       */
/* different values to the same variable.                */

static int matchingsconsistent( m1, m2 )
int m1;
int m2;
{
   int b1, e1;
   int b2, e2;
   int i, j;

   b1 = matchinglists [ m1 ]. matchingstart; 
   e1 = matchinglists [ m1 ]. matchingend; 
   b2 = matchinglists [ m2 ]. matchingstart; 
   e2 = matchinglists [ m2 ]. matchingend;

   for( i = b1; i < e1; i ++ ) 
   {
      for( j = b2; j < e2; j ++ ) 
      {
         if( assignments [i]. var == assignments [j]. var )
         {
            if( ! termsareequal( assignments [i]. term,
                                 assignments [j]. term ))
               return 0;
         }
      }
   }
   return 1;
}



/********************************************************************/
/*       Functions for matchings.                                   */
/********************************************************************/


/* Try to match term1 into term2, and collect the assignments to the */
/* variables, without checking if the assignments are consistent.    */
/* This is done later in matchingconsistent.                         */
/* The returnvalue is -1 on failure, or the index in matchinglists.  */

static int semimatch0 ( term1, term2 ) 
int term1;
int term2;
{
   int a; 
   int term1end;
   int firstassignment; 

   firstassignment = freeassignments;  
   term1end = termends [ term1 ];

   while( term1 < term1end )  
   {
      a = termspace [ term1 ++ ]; 
      if( a < 0 )
      {
         assignments [ freeassignments ]. var = a;
         assignments [ freeassignments ++ ]. term = term2;
         if( freeassignments >= MAXASSIGNMENTS - 3 )
             errorwithoutline( "assignments is full" );
         term2 = termends [ term2 ]; 
      }
      else
      {
         if( a != termspace [ term2 ++ ] )
         {
            freeassignments = firstassignment;
            return -1;
         }
      }
   next: 
      ;
   }
   matchinglists [ freematchinglists ]. matchingstart = firstassignment;
   matchinglists [ freematchinglists ]. matchingend = freeassignments;
   matchinglists [ freematchinglists ]. rest = -1; 

   if( freematchinglists >= MAXMATCHINGLISTS - 3 )
      errorwithoutline( "matchinglists is full" ); 

   return freematchinglists ++;
}
               

/* Try to construct a matching of term1 into term2, with the  */
/* additional constraint that the matching must be a          */
/* renaming. For the rest semimatch1 behaves the same as      */
/* makematchsnl.                                              */


static int semimatch1( term1, term2 ) 
int term1;
int term2;
{
   int a; 
   int term1end;
   int firstassignment; 

   firstassignment = freeassignments;  
   term1end = termends [ term1 ];

#if 0
   printstring( &outputfile, "semimatch1 ");
   printterm( &outputfile, term1 );
   printstring( &outputfile, ",  " ); 
   printterm( &outputfile, term2 );
   printstring( &outputfile, "\n" );
#endif

   if( term1end - term1 != termends [ term2 ] - term2 )
      return -1;

   while( term1 < term1end )  
   {
      a = termspace [ term1 ++ ]; 
      if( a < 0 )
      {
         if( termspace [ term2 ] >= 0 )
         {
            freeassignments = firstassignment; 
            return -1;
         }
         assignments [ freeassignments ]. var = a;
         assignments [ freeassignments ++ ]. term = term2;
         if( freeassignments >= MAXASSIGNMENTS - 3 )
            errorwithoutline( "assignments is full" );
         term2 = termends [ term2 ]; 
      }
      else
      {
         if( a != termspace [ term2 ++ ] )
         {
            freeassignments = firstassignment;
            return -1;
         }
      }
   next: 
      ;
   }
   matchinglists [ freematchinglists ]. matchingstart = firstassignment;
   matchinglists [ freematchinglists ]. matchingend = freeassignments;
   matchinglists [ freematchinglists ]. rest = -1; 

   if( freematchinglists >= MAXMATCHINGLISTS - 3 )
      errorwithoutline( "matchinglists is full" ); 

   return freematchinglists ++;
}


/* Try to match term1 into term2, possibly matching = into ==>.     */
/* If dir == 0, matching is from left to right.                     */
/* If dir == 1, and both term1 and term2 are constructed by         */
/* =, or ==>, then then matching is tried with the arguments        */
/* exchanged.                                                       */


static int semimatch2( term1, term2, dir )
int term1;
int term2;
int dir;
{
   int a;
   int t1, t1end;
   int term1end; 
   int firstassignment;

#if 0
   printstring( &outputfile, "semimatch2 " );
   printterm( &outputfile, term1 );
   printstring( &outputfile, ",  ");
   printterm( &outputfile, term2 );
   printstring( &outputfile, ", dir = " );
   printint( &outputfile, dir ); 
   printstring( &outputfile, "\n" );
#endif

   if( dir == 0 )
   {
      term1end = termends [ term1 ];

      if( termspace [ term1 ] == OP_NOT &&
          termspace [ term2 ] == OP_NOT )
      {
         term1 ++ ; 
         term2 ++ ;
      }
      if( termspace [ term1 ] == OP_EQUAL || 
          termspace [ term1 ] == OP_REWRITES )
      {
         if( termspace [ term2 ] == OP_EQUAL || 
             termspace [ term2 ] == OP_REWRITES ) 
         {
            term1 ++ ;
            term2 ++ ;
         }
         else
         {
            return -1;
         }
      } 

      firstassignment = freeassignments;

      while( term1 < term1end )
      {
         a = termspace [ term1 ++ ];
         if( a < 0 )
         {
            assignments [ freeassignments ]. var = a;
            assignments [ freeassignments ++ ]. term = term2;
            if( freeassignments >= MAXASSIGNMENTS - 3 )
               errorwithoutline( "assignments is full" );
            term2 = termends [ term2 ];
         }
         else
         {
            if( a != termspace [ term2 ++ ] )
            {
               freeassignments = firstassignment;
               return -1;
            }
         }
      }
   }
   else
   {
      if( termspace [ term1 ] == OP_NOT && termspace [ term2 ] == OP_NOT ) 
      {
         term1 ++ ;
         term2 ++ ;
      }
      if( termspace [ term1 ] != OP_EQUAL && 
          termspace [ term1 ] != OP_REWRITES ) 
         return -1;
      if( termspace [ term2 ] != OP_EQUAL &&
          termspace [ term2 ] != OP_REWRITES ) 
         return -1;
   
      firstassignment = freeassignments;    
      t1 = termends [ term1 + 1 ];
      t1end = termends [ t1 ];
      term2 ++ ;

      while( t1 < t1end )
      {
         a = termspace [ t1 ++ ];
         if( a < 0 )
         {
            assignments [ freeassignments ]. var = a;
            assignments [ freeassignments ++ ]. term = term2;  
            if( freeassignments >= MAXASSIGNMENTS - 3 ) 
               errorwithoutline( "assignments is full" );
            term2 = termends [ term2 ];
         }
         else
         {
            if( a != termspace [ term2 ++ ])
            {
               freeassignments = firstassignment;
               return -1;
            }
         }
      }

      t1 = term1 + 1;
      t1end = termends [ t1 ];

      while( t1 < t1end )
      {
         a = termspace [ t1 ++ ] ;
         if( a < 0 )
         {
            assignments [ freeassignments ]. var = a;
            assignments [ freeassignments ++ ]. term = term2;
             if( freeassignments >= MAXASSIGNMENTS - 3 )
               errorwithoutline( "assignments is full" );
            term2 = termends [ term2 ];
         }
         else 
         {
            if( a != termspace [ term2 ++ ] )
            {
               freeassignments = firstassignment;
               return -1;
            }
         }
      }
   }
   matchinglists [ freematchinglists ]. matchingstart = firstassignment;
   matchinglists [ freematchinglists ]. matchingend = freeassignments;
   matchinglists [ freematchinglists ]. rest = -1;
   
   if( freematchinglists >= MAXMATCHINGLISTS - 3 ) 
      errorwithoutline( "matchinglists is full" );
   
   return freematchinglists ++ ;
}

  


/***************************************************/
/* Semantic Tableau Style fashion backtracking.    */
/* This is only done when all else fails.          */
/* The parameter n is the depth.                   */
/* backtrack searches for a consistent path        */
/* through litmatchings[0..freelitmatchings -1].   */
/* matchinglist. It only backtracks on matchings   */
/* that contributed to the contradiction.          */
/***************************************************/

#define DEBUGBACKTRACK   0

static int backtrack( n )
int n;
{
   int best, mostvar;
   int i;
   int m;

#if DEBUGBACKTRACK 
   printstring( &outputfile, "\n\nentering backtrack on level ");
   printint( &outputfile, n );
   printstring( &outputfile, "\n");
#endif

#if DEBUGBACKTRACK == 2 
   printstring( &outputfile, "Semtab:\n\n");
   for( i = 0; i < n; i ++ ) 
   {
      printint( &outputfile, i );
      printstring( &outputfile, ",  used:" );
      printint( &outputfile, semtab[i]. conflict ); 
      printstring( &outputfile, ",  match: " );
      printmatching( &outputfile, semtab [i]. m );
      printstring( &outputfile, "\n" );
   }
   printstring( &outputfile, "\n" );
#endif

   if( n == freelitmatchings )
   {
      goto return1;
   }
  
   /* determine which is the best to start with.    */
   /* This is one which has nralternatives == 1,    */
   /* or otherwise the largest number of variables. */

   mostvar = -1;
   for( i = 0; i < freelitmatchings; i ++ ) 
   {
      if( litmatchings [i]. notinsemtab )
      {
         if( litmatchings [i]. nralternatives == 1 )
         {
            best = i;
            goto decided;
         }
         if( litmatchings [i]. nrvariables > mostvar )
         {
            best = i;
            mostvar = litmatchings [i]. nrvariables;
         }
      }
   }
decided:  

#if DEBUGBACKTRACK == 2 
   printstring( &outputfile, "extending ");
   printint( &outputfile, best );
   printstring( &outputfile, "\n");
#endif

   litmatchings [ best ]. notinsemtab = 0;
   for( m = litmatchings [ best ]. matchinglist;
        m >= 0; 
        m = matchinglists [m]. rest )
   {
      for( i = 0; i < n; i ++ )
      {
         if( ! matchingsconsistent( m, semtab [i]. m ))
         {
            semtab [i]. conflict = 1;
            goto conflict;
         }
      }
      semtab [n]. conflict = 0;
      semtab [n]. m = m;
      if( backtrack( n + 1 ))
         goto return1;
      if( ! semtab [n]. conflict )
      {
#if DEBUGBACKTRACK == 2 
         printint( &outputfile, n );
         printstring( &outputfile, " was not used\n");
#endif
         goto notused;
      }
   conflict: 
      ;
   }
notused:
   litmatchings [ best ]. notinsemtab = 1;

return0:
#if DEBUGBACKTRACK
   printstring( &outputfile, "failure on level ");
   printint( &outputfile, n );
   printstring( &outputfile, "\n");
#endif
   return 0;

return1:
#if DEBUGBACKTRACK 
   printstring( &outputfile, "success on level ");
   printint( &outputfile, n );
   printstring( &outputfile, "\n");
#endif
   return 1;
}



/******************************************************/
/*        return 1 if clause1 subsumes clause2.       */
/******************************************************/

#define DEBUGSUBSUMES 0

/* 0 is no information.                       */
/* 1 is only entry and exit information.      */
/* 2 gives full debugging information.        */


int subsumes( clause1, clause2 )
int clause1;
int clause2;
{
   /* these are references into termspace:  */ 

   int clause1start, clause2start;
   int clause1end, clause2end;
   int clause1len, clause2len;
   int ref1, ref2;
   int dir2;
   int m;
   int othermatch;

#if DEBUGSUBSUMES 
   printstring( &outputfile, "\nsubsumes: \n" ); 
   printclause( &outputfile, clause1 );
   printclause( &outputfile, clause2 );
   printstring( &outputfile, "?\n\n"); 
#endif

   subsentry ++ ; 
    
   clause1len = clauses [ clause1 + CLAUSE_LENGTH ];
   clause2len = clauses [ clause2 + CLAUSE_LENGTH ];

   if( clause1len > clause2len )
      goto return0;

   if( clauses [ clause1 + CLAUSE_WEIGHT ] > 
       clauses [ clause2 + CLAUSE_WEIGHT ] ) 
      goto return0;

   if( clauses [ clause1 + CLAUSE_DEPTH ] > 
       clauses [ clause2 + CLAUSE_DEPTH ] ) 
      goto return0;

   clause1start = clauses [ clause1 + CLAUSE_LITERALS ];
   clause2start = clauses [ clause2 + CLAUSE_LITERALS ]; 
   clause1end = clauses [ clause1 + CLAUSE_END ]; 
   clause2end = clauses [ clause2 + CLAUSE_END ];

   freeassignments = 0;
   freematchinglists = 0;

   switch( substype )
   {

   case SUBS_STANDARD:
   case0: 
      /* The usual type of subsumption.    */


      /* Try to match the literals.  */
      
      for( ref1 = clause1start; ref1 < clause1end; ref1 ++ )
      {
         for( ref2 = clause2start; ref2 < clause2end; ref2 ++ )
         {
            m = semimatch0( clauses [ ref1 ], clauses[ ref2 ] );
            if( m >= 0 )
            {
               litmatchings [ ref1 - clause1start ]. matchinglist = m;
               litmatchings [ ref1 - clause1start ]. firstfit = ref2;
               goto case0semimatch;
            }
         }
         goto return0;
      case0semimatch: 
         ;
      }

      /* At this point, we have been able to make a semi-matching.  */
      /* We now check their consistency.                            */

      subspropmatch ++ ; 

      for( ref1 = clause1start; ref1 < clause1end; ref1 ++ )
      {
         if( ! matchingconsistent( 
                  litmatchings [ ref1 - clause1start ]. matchinglist ))
         {
            for( ref2 = litmatchings [ ref1 - clause1start ]. firstfit + 1;
                 ref2 < clause2end; ref2 ++ )
            {
               m = semimatch0( clauses [ ref1 ], clauses [ ref2 ] );
               if( m >= 0 && matchingconsistent( m ))
               {
                  litmatchings [ ref1 - clause1start ]. matchinglist = m;
                  litmatchings [ ref1 - clause1start ]. firstfit = ref2;
                  goto case0match; 
               }
            }
            goto return0;
         }
      case0match: 
         ;
      }

      freelitmatchings = clause1len;

#if DEBUGSUBSUMES == 2  
      printstring( &outputfile, "The first matchings:\n");
      printlitmatchings( &outputfile, clause2start );
#endif

      /* Arrived here, we know that all matchings are internally consistent. */

      subslitmatch ++ ;

      for( ref1 = 0; ref1 < freelitmatchings; ref1 ++ )
      {
         for( ref2 = 0; ref2 < ref1; ref2 ++ )
         {  
            if( ! matchingsconsistent( 
                     litmatchings[ ref1 ]. matchinglist,
                     litmatchings[ ref2 ]. matchinglist ))
               goto case0allmatchings; 
         }
      }
      goto return1;

   case0allmatchings:

      /* Produce all alternative matchings.    */

      othermatch = 0;
      for( ref1 = clause1start; ref1 < clause1end; ref1 ++ )
      {
         litmatchings [ ref1 - clause1start ]. nralternatives = 1;
         for( ref2 = litmatchings [ ref1 - clause1start ]. firstfit + 1;
              ref2 < clause2end; ref2 ++ )
         {
            m = semimatch0( clauses [ ref1 ], clauses [ ref2 ] );
            if( m >= 0 && matchingconsistent( m ))
            {
               othermatch = 1;
               matchinglists [m]. rest = 
                  litmatchings [ ref1 - clause1start ]. matchinglist;
               litmatchings [ ref1 - clause1start ]. matchinglist = m;
               litmatchings [ ref1 - clause1start ]. nralternatives ++ ;
            }
         }
      }

      if( ! othermatch )
         goto return0; 

      /* We have been unlucky, and have to do the full backtracking. */

      subsfullsubs ++ ;

      for( ref1 = clause1start; ref1 < clause1end; ref1 ++ )
      {
         m = litmatchings [ ref1 - clause1start ]. matchinglist; 
         litmatchings [ ref1 - clause1start ]. nrvariables = 
            matchinglists [m]. matchingend - 
            matchinglists [m]. matchingstart;
         litmatchings [ ref1 - clause1start ]. notinsemtab = 1; 
      }

#if DEBUGSUBSUMES == 2 
      printstring( &outputfile, "\n\nAll matchings:\n\n" );
      printlitmatchings( &outputfile, clause2start ); 
#endif

#if DEBUGSUBSUMES 
   if( backtrack(0))
      goto return1;
   goto return0;
#else
   return backtrack( 0 ); 
#endif


   case SUBS_SNL1:                     
      /* The substitution has to be a renaming                      */
      /* on each literal separately, but not on the clause as a     */
      /* whole.                                                     */
  
      if( clause1len == 1 )
         goto case0;
 
      /* Try to match the literals.  */
      
      for( ref1 = clause1start; ref1 < clause1end; ref1 ++ )
      {
         for( ref2 = clause2start; ref2 < clause2end; ref2 ++ )
         {
            m = semimatch1( clauses [ ref1 ], clauses[ ref2 ] );
            if( m >= 0 )
            {
               litmatchings [ ref1 - clause1start ]. matchinglist = m;
               litmatchings [ ref1 - clause1start ]. firstfit = ref2;
               goto case1semimatch;
            }
         }
         goto return0;
      case1semimatch: 
         ;
      }

      /* At this point, we have been able to make a semi-matching.  */
      /* We now check their consistency.                            */

      subspropmatch ++ ; 

      for( ref1 = clause1start; ref1 < clause1end; ref1 ++ )
      {
         m = litmatchings [ ref1 - clause1start ]. matchinglist; 
         if( ! matchingconsistent( m ) || ! matchingrenaming( m ))
         {
            for( ref2 = litmatchings [ ref1 - clause1start ]. firstfit + 1;
                 ref2 < clause2end; ref2 ++ )
            {
               m = semimatch1( clauses [ ref1 ], clauses [ ref2 ] );
               if( m >= 0 && matchingconsistent( m ) && matchingrenaming( m ))
               {
                  litmatchings [ ref1 - clause1start ]. matchinglist = m;
                  litmatchings [ ref1 - clause1start ]. firstfit = ref2;
                  goto case1match; 
               }
            }
            goto return0;
         }
      case1match: 
         ;
      }

      freelitmatchings = clause1len;

#if DEBUGSUBSUMES == 2  
      printstring( &outputfile, "The first matchings:\n");
      printlitmatchings( &outputfile, clause2start );
#endif

      /* Arrived here, we know that all matchings are internally consistent. */

      subslitmatch ++ ;

      for( ref1 = 0; ref1 < freelitmatchings; ref1 ++ )
      {
         for( ref2 = 0; ref2 < ref1; ref2 ++ )
         {  
            /* No special conditions here. */

            if( ! matchingsconsistent( 
                     litmatchings[ ref1 ]. matchinglist,
                     litmatchings[ ref2 ]. matchinglist ))
               goto case1allmatchings; 
         }
      }
      goto return1;

   case1allmatchings:

      /* Produce all alternative matchings.    */

      othermatch = 0;
      for( ref1 = clause1start; ref1 < clause1end; ref1 ++ )
      {
         litmatchings [ ref1 - clause1start ]. nralternatives = 1;
         for( ref2 = litmatchings [ ref1 - clause1start ]. firstfit + 1;
              ref2 < clause2end; ref2 ++ )
         {
            m = semimatch1( clauses [ ref1 ], clauses [ ref2 ] );
            if( m >= 0 && matchingconsistent( m ) && matchingrenaming( m ))
            {
               othermatch = 1;
               matchinglists [m]. rest = 
                  litmatchings [ ref1 - clause1start ]. matchinglist;
               litmatchings [ ref1 - clause1start ]. matchinglist = m;
               litmatchings [ ref1 - clause1start ]. nralternatives ++ ;
            }
         }
      }

      if( ! othermatch )
         goto return0; 

      /* We have been unlucky, and have to do the full backtracking. */

      subsfullsubs ++ ;

      for( ref1 = clause1start; ref1 < clause1end; ref1 ++ )
      {
         m = litmatchings [ ref1 - clause1start ]. matchinglist; 
         litmatchings [ ref1 - clause1start ]. nrvariables = 
            matchinglists [m]. matchingend - 
            matchinglists [m]. matchingstart;
         litmatchings [ ref1 - clause1start ]. notinsemtab = 1; 
      }

#if DEBUGSUBSUMES == 2 
      printstring( &outputfile, "\n\nAll matchings:\n\n" );
      printlitmatchings( &outputfile, clause2start ); 
#endif

#if DEBUGSUBSUMES 
   if( backtrack(0))
      goto return1;
   goto return0;
#else
   return backtrack( 0 ); 
#endif


   case SUBS_EQREWR:
      /* SUBS_EQREWR does not distinghuish = from ==>, and matches  */
      /* t1 =/==> t2 into t2 =/==> t1.                              */

      /* Try to match the literals.  */
      
      for( ref1 = clause1start; ref1 < clause1end; ref1 ++ )
      {
         m = termspace [ clauses [ ref1 ]];
         if( m == OP_NOT ) 
            m = termspace [ clauses [ ref1 ] + 1 ];
         if( m == OP_EQUAL || m == OP_REWRITES ) 
            litmatchings [ ref1 - clause1start ]. supdir = 2;
         else
            litmatchings [ ref1 - clause1start ]. supdir = 1; 

         for( ref2 = clause2start; ref2 < clause2end; ref2 ++ )
         {
            m = semimatch2( clauses [ ref1 ], clauses[ ref2 ], 0 );
            if( m >= 0 )
            {
               litmatchings [ ref1 - clause1start ]. matchinglist = m;
               litmatchings [ ref1 - clause1start ]. firstfit = ref2;
               litmatchings [ ref1 - clause1start ]. firstdir = 0; 
               goto case2semimatch;
            }
         }

         if( litmatchings [ ref1 - clause1start ]. supdir == 2 ) 
         {
            for( ref2 = clause2start; ref2 < clause2end; ref2 ++ )
            {
               m = semimatch2( clauses [ ref1 ], clauses [ ref2 ], 1 );
               if( m >= 0 )
               {
                  litmatchings [ ref1 - clause1start ]. matchinglist = m;
                  litmatchings [ ref1 - clause1start ]. firstfit = ref2;
                  litmatchings [ ref1 - clause1start ]. firstdir = 1;
                  goto case2semimatch;
               }
            }
         }
         goto return0;
      case2semimatch: 
         ;
      }

      /* At this point, we have been able to make a semi-matching.  */
      /* We now check their consistency.                            */

      subspropmatch ++ ; 

      for( ref1 = clause1start; ref1 < clause1end; ref1 ++ )
      {
         if( ! matchingconsistent( 
                  litmatchings [ ref1 - clause1start ]. matchinglist ))
         {
            ref2 = litmatchings [ ref1 - clause1start ]. firstfit + 1;
            dir2 = litmatchings [ ref1 - clause1start ]. firstdir;
            if( ref2 == clause2end )
            {
               ref2 = clause2start;
               dir2 ++ ;
            }
            while( dir2 < litmatchings [ ref1 - clause1start ]. supdir )  
            {
               m = semimatch2( clauses [ ref1 ], clauses [ ref2 ], dir2 );
               if( m >= 0 && matchingconsistent( m ))
               {
                  litmatchings [ ref1 - clause1start ]. matchinglist = m;
                  litmatchings [ ref1 - clause1start ]. firstfit = ref2;
                  litmatchings [ ref1 - clause1start ]. firstdir = dir2; 
                  goto case2match; 
               }
               ref2 ++ ;
               if( ref2 == clause2end )
               {
                  ref2 = clause2start;
                  dir2 ++ ;
               }
            }
            goto return0;
         }
      case2match: 
         ;
      }

      freelitmatchings = clause1len;

#if DEBUGSUBSUMES == 2  
      printstring( &outputfile, "The first matchings:\n");
      printlitmatchings( &outputfile, clause2start );
#endif

      /* Arrived here, we know that all matchings are internally consistent. */

      subslitmatch ++ ;

      for( ref1 = 0; ref1 < freelitmatchings; ref1 ++ )
      {
         for( ref2 = 0; ref2 < ref1; ref2 ++ )
         {  
            if( ! matchingsconsistent( 
                     litmatchings[ ref1 ]. matchinglist,
                     litmatchings[ ref2 ]. matchinglist ))
               goto case2allmatchings; 
         }
      }
      goto return1;

   case2allmatchings:

      /* Produce all alternative matchings.    */

      othermatch = 0;
      for( ref1 = clause1start; ref1 < clause1end; ref1 ++ )
      {
         litmatchings [ ref1 - clause1start ]. nralternatives = 1;

         ref2 = litmatchings [ ref1 - clause1start ]. firstfit + 1;
         dir2 = litmatchings [ ref1 - clause1start ]. firstdir;
         if( ref2 == clause2end )
         {
            ref2 = clause2start;
            dir2 ++ ;
         }
         while( dir2 < litmatchings [ ref1 - clause1start ]. supdir )
         {
            m = semimatch2( clauses [ ref1 ], clauses [ ref2 ], dir2 );
            if( m >= 0 && matchingconsistent( m ))
            {
               othermatch = 1;
               matchinglists [m]. rest =
                  litmatchings [ ref1 - clause1start ]. matchinglist;
               litmatchings [ ref1 - clause1start ]. matchinglist = m;
               litmatchings [ ref1 - clause1start ]. nralternatives ++ ;
            }
            ref2 ++ ;
            if( ref2 == clause2end )
            {
               ref2 = clause2start;
               dir2 ++ ;
            }
         }
      }

      if( ! othermatch )
         goto return0; 

      /* We have been unlucky, and have to do the full backtracking. */

      subsfullsubs ++ ;

      for( ref1 = clause1start; ref1 < clause1end; ref1 ++ )
      {
         m = litmatchings [ ref1 - clause1start ]. matchinglist; 
         litmatchings [ ref1 - clause1start ]. nrvariables = 
            matchinglists [m]. matchingend - 
            matchinglists [m]. matchingstart;
         litmatchings [ ref1 - clause1start ]. notinsemtab = 1; 
      }

#if DEBUGSUBSUMES == 2 
      printstring( &outputfile, "\n\nAll matchings:\n\n" );
      printlitmatchings( &outputfile, clause2start ); 
#endif

#if DEBUGSUBSUMES 
   if( backtrack(0))
      goto return1;
   goto return0;
#else
   return backtrack( 0 ); 
#endif


   default:
      internalproblem( "unknown type of subsumption" );
      return 0;
   }

return0:
#if DEBUGSUBSUMES
   printstring( &outputfile, " ==> 0 \n" );
#endif
   return 0;

return1:
#if DEBUGSUBSUMES
   printstring( &outputfile, " ==> 1 \n" );
#endif
   return 1;
} 


/****************************************************************/
/*                   Forwardsubsumes.                           */
/****************************************************************/

/* clear the n-th bit in the filter of clause:        */

static void clearnthbit( clause, n )
int clause;
int n;
{
   int bit, position;

   position = n / ( 8 * sizeof(int) );
   bit = 1 << ( n % ( 8 * sizeof(int) ));
   bit = -1 ^ bit;

   clauses [ clause + CLAUSE_FILT1 + position ] =
      clauses [ clause + CLAUSE_FILT1 + position ] & bit;
}


/* construct the contribution of term to the subsumption filter. */
/* context is the bitnr belonging to the parent term.            */
/* maxhash is the number of available bits in the filter.        */


static void makefpa( clause, context, term, maxhash )
int clause;
int context;
int term;
int maxhash;
{
   int i, arity;
   int f;

   f = termspace [ term ];
   if( f >= 0 )
   {
      if( f == OP_REWRITES ) 
         f = OP_EQUAL;

      context = 31 * context + f;
      context = context % maxhash;
      while( context < 0 )
         context += maxhash;
      if( context >= maxhash ) 
      {
         internalproblem( "problems with hash!\n\n");
         exit(0);
      }
      clearnthbit( clause, context );

      arity = arities [f];
      term ++ ;
      context *= 7;

      if( f != OP_EQUAL || substype != SUBS_EQREWR )
      {
         for( i = 0; i < arity; i ++ )
         {
            makefpa( clause, context, term, maxhash );
            context ++ ;
            term = termends [ term ];
         }
      }
      else
      {
         for( i = 0; i < arity; i ++ )
         {
            makefpa( clause, context, term, maxhash );
            term = termends [ term ];
         }
      }
   }
}


/* Make the filters for a clause:   */

void makefilter( clause )
int clause;
{
   int c;
   int maxhash;
   int term;

   clauses [ clause + CLAUSE_FILT1 ] = -1;
   clauses [ clause + CLAUSE_FILT2 ] = -1;
   clauses [ clause + CLAUSE_FILT3 ] = -1;
   clauses [ clause + CLAUSE_FILT4 ] = -1;
   clauses [ clause + CLAUSE_FILT5 ] = -1;
   clauses [ clause + CLAUSE_FILT6 ] = -1;
   clauses [ clause + CLAUSE_FILT7 ] = -1;
   clauses [ clause + CLAUSE_FILT8 ] = -1;

   maxhash = 8 * 8 * ( sizeof( int ));

   for( c = clauses [ clause + CLAUSE_LITERALS ];
        c < clauses [ clause + CLAUSE_END ];
        c ++ )
   {
      term = clauses [c];
      makefpa( clause, 0, term, maxhash );
   }
}


/* return 1 if clause is forward subsumed.      */

#define BITOK(A,B) ( ((A) | (B)) == -1 )

/* return 1 if one of the clause in list subsumes clause. */

static int listsubsumes( list, clause )
register int list;
int clause;
{
   register int filt1, filt2, filt3, filt4;
   int filt5, filt6, filt7, filt8;

   filt1 = -1 ^ clauses [ clause + CLAUSE_FILT1 ];
   filt2 = -1 ^ clauses [ clause + CLAUSE_FILT2 ];
   filt3 = -1 ^ clauses [ clause + CLAUSE_FILT3 ];
   filt4 = -1 ^ clauses [ clause + CLAUSE_FILT4 ];
   filt5 = -1 ^ clauses [ clause + CLAUSE_FILT5 ];
   filt6 = -1 ^ clauses [ clause + CLAUSE_FILT6 ];
   filt7 = -1 ^ clauses [ clause + CLAUSE_FILT7 ];
   filt8 = -1 ^ clauses [ clause + CLAUSE_FILT8 ];

   while( list != EMPTYCLAUSELIST )
   {
      if( 
#if 1
          BITOK( clauses [ list + CLAUSE_FILT1 ], filt1 ) &&
          BITOK( clauses [ list + CLAUSE_FILT2 ], filt2 ) &&
          BITOK( clauses [ list + CLAUSE_FILT3 ], filt3 ) &&
          BITOK( clauses [ list + CLAUSE_FILT4 ], filt4 ) &&
          BITOK( clauses [ list + CLAUSE_FILT5 ], filt5 ) &&
          BITOK( clauses [ list + CLAUSE_FILT6 ], filt6 ) &&
          BITOK( clauses [ list + CLAUSE_FILT7 ], filt7 ) &&
          BITOK( clauses [ list + CLAUSE_FILT8 ], filt8 ) &&
#endif
          subsumes ( list, clause ) && list != clause )
      {
         return 1;
      }

      list = clauses [ list + CLAUSE_NEXTINLIST ];
   }
   return 0;
}



/* Return 1 if the filters of clause1 and clause are OK.   */

static int subsfilter( clause1, clause2 )
int clause1;
int clause2;
{
   if( ! BITOK( clauses [ clause1 + CLAUSE_FILT1 ],
             -1 ^ clauses [ clause2 + CLAUSE_FILT1 ] ))
      return 0;
   if( ! BITOK( clauses [ clause1 + CLAUSE_FILT2 ],
              -1 ^ clauses [ clause2 + CLAUSE_FILT2 ] ))
      return 0;
   if( ! BITOK( clauses [ clause1 + CLAUSE_FILT3 ],
              -1 ^ clauses [ clause2 + CLAUSE_FILT3 ] ))
      return 0;
   if( ! BITOK( clauses [ clause1 + CLAUSE_FILT4 ],
              -1 ^ clauses [ clause2 + CLAUSE_FILT4 ] ))
      return 0;
   if( ! BITOK( clauses [ clause1 + CLAUSE_FILT5 ],
              -1 ^ clauses [ clause2 + CLAUSE_FILT5 ] ))
      return 0;
   if( ! BITOK( clauses [ clause1 + CLAUSE_FILT6 ],
              -1 ^ clauses [ clause2 + CLAUSE_FILT6 ] ))
      return 0;
   if( ! BITOK( clauses [ clause1 + CLAUSE_FILT7 ],
              -1 ^ clauses [ clause2 + CLAUSE_FILT7 ] ))
      return 0;
   if( ! BITOK( clauses [ clause1 + CLAUSE_FILT8 ],
              -1 ^ clauses [ clause2 + CLAUSE_FILT8 ] ))
      return 0;
   return 1;
}


static int subs_clause;
static jmp_buf subs_yes;

static void forwardsubsumed2( clause, literal, term )
int clause;
int literal;
int term;
{
   /* In the case of backward subsumption, the clause itself will be */
   /*retrieved.                                                      */

   if( clause == subs_clause )
      return;

#if 0
   printstring( &outputfile, "Retrieved Possible Subsumer:\n");
   printclause( &outputfile, clause );
   printstring( &outputfile, "\n" );
#endif

   /* This doesn't seem to help:
   if( clauses [ clause + CLAUSE_LENGTH ] == 1 && backwardsubs == 1 )
      longjmp( subs_yes, 1 ); */

   if( ! subsfilter( clause, subs_clause ))
      return;

   if( subsumes( clause, subs_clause ))
      longjmp( subs_yes, 1 );

   return;
}




/* Check if clause is forward subsumed.                    */


int forwardsubsumed( clause )
int clause;
{
   int i;

#if 0
   printstring( &outputfile, "Subsumption test for: ");
   printclause( &outputfile, clause );
#endif

   switch( backwardsubs )
   {
   case 0:
      if( listsubsumes( listunfactored, clause ))
         return 1;
      if( listsubsumes( listinuse, clause ))
         return 1;
      if( listsubsumes( listclauses, clause ))
         return 1;
      return 0;
   case 1:
      subs_clause = clause;
      if( setjmp( subs_yes ))
      {
         return 1;
      }

      if( substype != SUBS_EQREWR )
      {
         for( i = clauses [ clause + CLAUSE_LITERALS ];
              i < clauses [ clause + CLAUSE_END ];
              i ++ )
         {
            retrievefromtree( treeforwsubs, clauses [i], forwardsubsumed2 );
         }
      }
      else
      {
         for( i = clauses [ clause + CLAUSE_LITERALS ];
              i < clauses [ clause + CLAUSE_END ];
              i ++ )
         {
            retrievefromtreeeqrewr( treeforwsubs, clauses [i], 
                                    forwardsubsumed2 );
         }
      }
      return 0;
   default:
      internalproblem( "improper value for backwardsubs" );
   }
}


int backwardsubsumed( clause )
int clause;
{
   return forwardsubsumed( clause );
}
 

/* Return the best literal of literals for the subsumption check. */
/* It is the largest with the smallest amount of variables.       */
/****** Should be insensitive to sorting of the clause. Improve it. ***/

static int termsimpcompl( t )
{
   return termends [t] - t;
}

static int bestsubsliteral( clause )
int clause;
{
   int best, i;
   int bestcomp, bestvarnr;
   int icomp, ivarnr;

   bestcomp = -1;
   for( i = clauses [ clause + CLAUSE_LITERALS ]; 
        i < clauses [ clause + CLAUSE_END ]; 
        i ++ )
   {
      icomp = termsimpcompl( clauses [i] );
      ivarnr = termnrvars( clauses [i] );
      if( icomp > bestcomp )
      {
         best = i;
         bestcomp = icomp;
         bestvarnr = ivarnr;
      }
      else
      {
         if( icomp == bestcomp && ivarnr < bestvarnr )
         {
            best = i;
            bestcomp = icomp;
            bestvarnr = ivarnr;
         }
      }
   }
   return best;
}


/* add or delete the most unlikely literal from treeforwsubst   */
/* if backwardsubs = 1.                                         */

void dt_subs( clause, adding )
int clause;
int adding;
{
   int best;

#if 0
   if( adding )
      printstring( &outputfile, "Adding Literal for: ");
   else
      printstring( &outputfile, "Deleting Literal for: ");
   printclause( &outputfile, clause );
#endif

   if( backwardsubs == 1 && clauses [ clause + CLAUSE_LENGTH ] > 0 )
   {
      best = bestsubsliteral( clause );
 
#if 0
      printstring( &outputfile, "Best: ");
      printterm( &outputfile, clauses [ best ] );
      printstring( &outputfile, "\n" ); 
#endif 
      if( adding )
         addtotree( &treeforwsubs, clause, best, clauses [ best ] );
      else
         deletefromtree( &treeforwsubs, clause, best, clauses [ best ] );
   }
}

 
