

/***********************************************************/
/* file unify.c contains unification and matching          */
/* together with functions for the handling of             */
/* discrimination trees.                                   */
/***********************************************************/

#define CLASS extern
#include "basics.h"
#include "extfunc.h"
#include "resolution.h"


/******************************************************************/
/*     Unification and Substitution.                              */
/******************************************************************/



#if 1

void printsubst( f )
bliksemfile *f;
{
   int i;

   printstring( f, "Substitution:\n\n" );

   for( i = 0; i < freesubst; i ++ ) 
   {
      printvariable( f, subst [i]. var ); 
      printstring( f, "[" ); 
      printint( f, subst [i]. varorig );
      printstring( f, "] := " );

      printterm( f, subst [i]. term );
      printstring( f, "[" );
      printint( f, subst [i]. termorig );
      printstring( f, "]\n" );
   }
   printstring( f, "\n\n" );
   for( i = 0; i < freerenaming; i ++ ) 
   {
      printvariable( f, renaming [i]. var );
      printstring( f, "[" );
      printint( f, renaming [i]. varorig );
      printstring( f, "] := " );
      printvariable( f, - ( 1 + i ));
      printstring( f, "\n" ); 
   }
   printstring( f, "\n\n" ); 
}

#endif


/* The occurs check:  */

static int occurscheck ( var, varorig, term, termorig )
int var;
int varorig;
int term;
int termorig;
{
   int f, i;
   int end;

   end = termends [ term ]; 
   while( term < end )
   {
      f = termspace [ term ++ ];
      if( f < 0 )
      {
         if( f == var && varorig == termorig )
            return 1;

         for( i = 0; i < freesubst; i ++ ) 
         {
            if( subst [i]. var == f && subst [i]. varorig == termorig ) 
            {
               if( occurscheck ( var, varorig, 
                                 subst [i]. term, subst [i]. termorig ))
                  return 1;
               goto next; 
            }
         }
      }
   next:
      ; 
   }
   return 0;
}


/* unify ( term1, orig1 ) with ( term2, orig2 ).    */
/* return 1 on success, and produce a substitution. */

int unify ( term1, orig1, term2, orig2 ) 
int term1;
int orig1;
int term2;
int orig2;
{
   int i;
   int f1, f2; 
   int endofterm1;

   endofterm1 = termends [ term1 ];
   while( term1 < endofterm1 )
   {
      f1 = termspace [ term1 ]; 
      if( f1 < 0 )
      {
         for( i = 0; i < freesubst; i ++ ) 
         {
            if( subst [i]. var == f1 && subst [i]. varorig == orig1 )
            {
               if( ! unify ( 
                     subst [i]. term, subst [i]. termorig,  
                     term2, orig2 )) 
                  return 0; 
               term1 ++ ; 
               term2 = termends [ term2 ]; 
               goto next; 
            }
         }
      }

      f2 = termspace [ term2 ]; 
      if( f2 < 0 ) 
      {
         for( i = 0; i < freesubst; i ++ ) 
         {
            if( subst [i]. var == f2 && subst [i]. varorig == orig2 )
            {
               if( ! unify (
                     term1, orig1, 
                     subst [i]. term, subst [i]. termorig )) 
                  return 0; 
               term2 ++ ; 
               term1 = termends [ term1 ]; 
               goto next;
            }
         }
      }

      /* both terms are now expanded under the existing substitution. */

      if( f1 < 0 )
      {
         if( f2 < 0 )
         {
            if( f1 != f2 || orig1 != orig2 )
            {
               if( freesubst >= MAXSUBST - 3 ) 
                  errorwithoutline( "substitution is too large" ); 

               subst [ freesubst ]. var = f1;
               subst [ freesubst ]. varorig = orig1;
               subst [ freesubst ]. term = term2;
               subst [ freesubst ]. termorig = orig2;
               freesubst ++ ;
            }
            term1 ++ ; term2 ++ ; 
            goto next; 
         }
         else
         {
            if( occurscheck ( f1, orig1, term2, orig2 ))
               return 0;

            if( freesubst >= MAXSUBST - 3 ) 
               errorwithoutline( "substitution is too large" );  

            subst [ freesubst ]. var = f1;
            subst [ freesubst ]. varorig = orig1;
            subst [ freesubst ]. term = term2;
            subst [ freesubst ]. termorig = orig2;
            freesubst ++ ; 

            term1 ++ ; 
            term2 = termends [ term2 ]; 
         }
      }
      else
      {
         if( f2 < 0 )
         {
            if( occurscheck ( f2, orig2, term1, orig1 ))
               return 0;       
       
            if( freesubst >= MAXSUBST - 3 ) 
               errorwithoutline( "substitution is too large" ); 

            subst [ freesubst ]. var = f2;
            subst [ freesubst ]. varorig = orig2;
            subst [ freesubst ]. term = term1;
            subst [ freesubst ]. termorig = orig1;
            freesubst ++ ; 

            term2 ++ ;
            term1 = termends [ term1 ]; 
         }
         else
         {
            if( f1 != f2 )
               return 0;
            term1 ++ ; term2 ++ ; 
         }
      }
   next:
      ;
   }
   return 1;
}

 

/* construct the result of (term, termorig) under the existing */
/* substitution, rename the variables in the result.           */


static void ressubst2( term, termorig )
int term;
int termorig;
{
   int f;
   int i;
   int end;
 
   end = termends [ term ];
   while( term < end ) 
   {
      while( freetermspace >= maxtermspace10 ) 
         more_termspace( ); 

      f = termspace[ term ++ ];
      if( f >= 0 )
         termspace [ freetermspace ++ ] = f;
      else
      {
         for( i = 0; i < freesubst; i ++ ) 
         {
            if( subst [i]. var == f && subst [i]. varorig == termorig )
            {
               ressubst2( subst [i]. term, subst [i]. termorig );
               goto next;
            }
         }
         for( i = 0; i < freerenaming; i ++ ) 
         {
            if( renaming [i]. var == f && renaming [i]. varorig == termorig )
            {
               termspace [ freetermspace ++ ] = - ( 1 + i);
               goto next;
            }
         }
         if( freerenaming >= MAXNRVARS - 3 )
            errorwithoutline( "renaming is full" ); 

         termspace [ freetermspace ++ ] = - ( 1 + freerenaming );

         renaming [ freerenaming ]. var = f;
         renaming [ freerenaming ]. varorig = termorig;
         freerenaming ++ ; 
      }
   next:
      ;
   }
}


/* Make a substitution.              */ 
/* The termends are not filled in.   */


int ressubst( term, termorig )
int term;
int termorig;
{
   int res;

   res = freetermspace;
   ressubst2( term, termorig );
   return res;
}


/*******************************************************************/
/* Matching and substitutions from matchings.                      */
/* Used for demodulation and simplifying resolution.               */
/*******************************************************************/


#if 1

void printmatching( f )
bliksemfile *f;
{
   int i;

   printstring( f, "Matching:\n\n" );

   for( i = 0; i < freematching; i ++ ) 
   {
      printvariable( f, matching [i]. var );
      printstring( f, " := " );
      printterm( f, matching [i]. term );  
      printstring( f, "\n"); 
   }
   printstring( f, "\n\n" ); 
}


#endif 


/* match term1 into term2. Return 1 on succes and store the matching   */
/* in the array matching [].                                           */


int match ( term1, term2 ) 
int term1;
int term2;
{
   int i;
   int f; 
   int term1end;

   term1end = termends [ term1 ];
   while( term1 < term1end )  
   {
      f = termspace [ term1 ++ ]; 
      if( f < 0 )
      {
         for( i = 0; i < freematching; i ++ ) 
         {
            if( matching [i]. var == f )
            {
               if( ! termequal( matching [i]. term, term2 ))
                  return 0;

               term2 = termends [ term2 ]; 
               goto next;
            }
         }
         matching [ freematching ]. var = f;
         matching [ freematching ]. term = term2;
         freematching ++ ;
         if( freematching >= MAXMATCHING - 3 )
            errorwithoutline( "matching is full" );
         term2 = termends [ term2 ]; 
      }
      else
      {
         if( f != termspace [ term2 ++ ] )
            return 0;
      }
   next: 
      ;
   }
   return 1; 
}
              

 
static void matchsubst2( term ) 
int term;
{
   int f;
   int i;
   int end;
   int t;
 
   end = termends [ term ];

   while( term < end )
   {
      f = termspace [ term ++ ];
      if( f < 0 )
      {
         for( i = 0; i < freematching; i ++ )
         {
            if( matching [i]. var == f )
            {
               for( t = matching [i]. term; 
                    t < termends [ matching [i]. term ]; 
                    t ++ )
               {
                  termspace [ freetermspace ++ ] = termspace [ t ];
                  while( freetermspace >= maxtermspace10 )
                     more_termspace( ); 
               }
               goto next;               
            }
         }
         internalproblem( "problems in matchsubst2 with rhs" );
      }
      else
      {
         while( freetermspace >= maxtermspace10 )
            more_termspace( ); 
         termspace [ freetermspace ++ ] = f;
      }
   next:
      ;
   }
}
            
 
/* Substitute term using the matching.    */

int matchsubst( term )
int term;
{
   int res;
 
   res = freetermspace;
   matchsubst2( term );
   return res;
}



/******************************************************************/
/* functions for the handling of discrimination trees.            */
/* At this moment we can only look for generalizations. So we     */
/* can use them for forward subsumption, forward demodulation,    */
/* and forward simplifying resolution.                            */
/******************************************************************/


/* This function also checks the format of the tree.    */ 

static void printtree2( f, tree, depth, nrvars )
bliksemfile *f;
int tree;
int depth;
int nrvars; 
{
   int i;
   int nrvars2;

   if( tree == DT_NIL ) 
      return;

   switch( clauses [ tree ] )
   {
   case DT_REC:
      while( tree != DT_NIL ) 
      {
         nrvars2 = nrvars; 
         if( clauses [ tree ] != DT_REC ) 
            internalproblem( "clauses [ tree ] != DT_REC" );

         for( i = 0; i < depth; i ++ )
            printstring( f, "   " );
         if( clauses [ tree + 1 ] >= 0 ) 
            printoperator( f, clauses [ tree + 1 ] ); 
         else
         {
            printvariable( f, clauses [ tree + 1 ] );
            if( -clauses [ tree + 1 ] - 1 > nrvars ) 
               internalproblem( "violation of normality" );
            if( -clauses [ tree + 1 ] - 1 == nrvars )
               nrvars2 = nrvars + 1;
         }
         printstring( f, "\n" );

         /* A forward reference cannot be DT_NIL. */

         if( clauses [ tree + 2 ] == DT_NIL ) 
         {
            internalproblem( "clauses [ tree + 2 ] != DT_NIL" );
            exit(0);
         }
         printtree2( f, clauses [ tree + 2 ], depth + 1, nrvars2 );
         tree = clauses [ tree + 3 ]; 
      }
      return; 
   case DT_SUCC:
      while( tree != DT_NIL ) 
      {
         if( clauses [ tree ] != DT_SUCC ) 
         {
            internalproblem( "clauses [ tree ] != DT_SUCC" );
            exit(0);
         }
         printstring( f, "Reference{ " );
         for( i = tree + 2; clauses [i] < 0; i ++ ) 
         {
            printvariable( f, clauses [i] );
            printstring( f, " " );
         }

         /* The number of variables in the renaming should */
         /* be equal to the number of variables in the     */
         /* term in the tree.                              */

         if( i - 2 - tree != nrvars ) 
         {
            internalproblem( "i - 2 - tree != nrvars" );
            exit(0);
         }
         printstring( f, "}:\n" );
         printclause( f, clauses [i++] );
         printstring( f, "literal:  "); 
         printterm( f, clauses [ clauses [i++]] ); 
         printstring( f, ",    term:  " ); 
         printterm( f, clauses [i++] );
         printstring( f, "\n" ); 
 
         tree = clauses [ tree + 1 ];
      }
      return;
   default:
      internalproblem( "reached default" );
      exit(0);
   }
}
         

/* Print a discrimination tree, and do a number of consistency checks on */
/* it.                                                                   */

void printtree( f, tree )
bliksemfile *f; 
int tree;
{
   printstring( f, "Discrimination tree:\n\n");
   printtree2( f, tree, 0, 0 ); 
   printstring( f, "\n\n" ); 
}


#if 0

/*******************************************************/
/*   Functions used for collecting statistics on       */
/*   the discrimination trees.                         */
/*******************************************************/

#define MAXALTERN      5000 

static int nodeall;
static int nodebacktrack;
static int nodevar [ MAXALTERN ];
static int nodenonvar [ MAXALTERN ];

static int nodesol; 
static int nodesolutions [ MAXALTERN ];


static void clearstats( ) 
{
   int i;

   for( i = 0; i < MAXALTERN; i ++ )
   {
      nodevar [i] = 0;
      nodenonvar [i] = 0;

      nodesolutions [i] = 0;
   }
   nodeall = 0;
   nodebacktrack = 0;
   nodesol = 0;
}



/* Count the alternativies in the node node, and add these to  */
/* the statistics.                                             */

static void countaltern( node )
int node;
{
   int vars, nonvars;

   nodeall ++ ;

   vars = 0;
   nonvars = 0;
   while( node != DT_NIL )
   {
      if( clauses [ node + 1 ] >= 0 )
         nonvars ++ ;
      else
         vars ++ ;   
      node = clauses [ node + 3 ];
   }

   if( vars + nonvars >= MAXALTERN - 3 ) 
   {
      internalproblem( "vars + nonvars >= MAXALTERN - 3" );
      exit(0);
   }

   nodevar [ vars ] ++ ;
   nodenonvar [ nonvars ] ++ ;

   if( vars >= 2 || ( vars == 1 && nonvars > 0 ) )
      nodebacktrack ++ ;
}


/* Count the statistics on the solution node node.   */

static void countsolu( node )
int node;
{
   int nrsol;

   nodesol ++ ;

   nrsol = 0;
   while( node != DT_NIL ) 
   {
      nrsol ++ ;
      node = clauses [ node + 1 ];
   }

   if( nrsol > MAXALTERN - 3 )
   {
      internalproblem( "nrsol > MAXALTERN - 3" );
      exit(0);
   }
   nodesolutions [ nrsol ] ++ ;    
}


/* print the statistics on the trees.     */

void printstats( )
{

   int column;
   int i;
   int totalprod;
   
   printf("Statistics:\n\n");

   printf("nodes: %d\n", nodeall );
   printf("backtrack: %d\n", nodebacktrack );

   printf("var: ");
   column = 0;
   for( i = 0; i < MAXALTERN; i ++ )
   {
      if( nodevar [i] > 0 )
      {
         printf("[ %d, %d ] ", i, nodevar [i] );
         column ++ ;
         if( column % 3 == 0 ) printf("\n   ");
      }
   }
   printf("\n");

   printf("nonvar: ");
   column = 0;
   totalprod = 0;
   for( i = 0; i < MAXALTERN; i ++ )
   {
      if( nodenonvar [i] > 0 )
      {
         printf("[ %d, %d, %d ] ", i, nodenonvar [i], i * nodenonvar [i] );
         column ++ ;
         totalprod += i * nodenonvar [i];
         if( column % 3 == 0 ) printf("\n   ");
      }
   }
   printf("\ntotalproduct: %d\n", totalprod );
   printf("\n");

   printf("sol: %d\n", nodesol );
   printf("solutions: ");
   column = 0;
   totalprod = 0;
   for( i = 0; i < MAXALTERN; i ++ )
   {
      if( nodesolutions [i] > 0 )
      {
         printf("[ %d, %d, %d ] ", i, nodesolutions [i], 
                                   i * nodesolutions [i] );
         totalprod += i * nodesolutions [i];
         column ++ ;
         if( column % 3 == 0 ) printf("\n   ");
      }
   }
   printf("\ntotalproduct: %d\n", totalprod ); 
   printf("\n\n");
}



static void counttree2( tree )
int tree;
{

   if ( tree == DT_NIL ) 
   {
      countaltern( tree ); 
      return; 
   }

   switch( clauses [ tree ] )  
   {
   case DT_REC:
      countaltern( tree );       
      while( tree != DT_NIL )
      {   
         if( clauses [ tree ] != DT_REC ) 
         {
            internalproblem( "clauses [ tree ] != DT_REC" );
            exit(0);
         }
         counttree2( clauses [ tree + 2 ] );
         tree = clauses [ tree + 3 ];
      }
      break;
   case DT_SUCC:
      /* countsolu( tree ); */
      while( tree != DT_NIL )
      {
         if( clauses [ tree ] != DT_SUCC )
         {
            internalproblem( "clauses [ tree ] != DT_SUCC" );
            exit(0);
         }
         tree = clauses [ tree + 1 ];
      }
      break;
   default:
      internalproblem( "reached the default" );
      exit(0);
   }   
}


/* collect statistics on the tree. This is useful in order to do        */
/* optimizations. One cannot do benchmarking without these statistics.  */

void counttree( tree )
int tree;
{
   clearstats( ); 

   counttree2( tree );

   printstring( &outputfile, "Giving statistics of tree:\n" );
   printstats( );
   printstring( &outputfile, "\n\n" );
}


#endif

/* add a term to *tree. clause is the clause from which term comes.  */
/* clauses [ literal ] contains the literal in the term. So term     */
/* should be a subterm of clauses [ literal ].                       */
 
void addtotree( tree, clause, literal, term )
int *tree;
int clause;
int literal;
int term;
{
   int f;
   int i;
   int end;
   int termind;

   if( termends [ term ] - term > 10000 )
      return;

   while( freeclauses + 5 * ( termends [ term ] - term ) + 100
       >= maxclauses10 ) 
   {
      more_clauses( ); 
   }

   freematching = 0; 
   end = termends [ term ]; 
   
   for( termind = term; termind < end; termind ++ ) 
   {
      f = termspace [ termind ];
      if( f < 0 )
      {
         for( i = 0; i < freematching; i ++ )
         {
            if( matching [i]. var == f )
               goto varfound;
         }
         if( freematching >= MAXMATCHING - 3 ) 
            errorwithoutline( "matching is full" ); 
         matching [ freematching ++ ]. var = f;
      varfound:
         f = -i - 1;
      }
      while( *tree != DT_NIL ) 
      {
         if( clauses [ *tree + 1 ] == f )
            goto opfound;
         tree = &clauses [ *tree + 3 ];
      }
#if 1 
      if( freeclauses >= maxclauses10 ) 
         errorwithoutline( "clauses is full and cannot be extended from here" ); 
#endif
      *tree = freeclauses;
      clauses [ freeclauses ++ ] = DT_REC;
      clauses [ freeclauses ++ ] = f;
      clauses [ freeclauses ++ ] = DT_NIL;
      clauses [ freeclauses ++ ] = DT_NIL;
   opfound:
      tree = &clauses [ *tree + 2 ]; 
   }
 
   while( *tree != DT_NIL ) 
      tree = &clauses [ *tree + 1 ];
#if 1
   if( freeclauses + freematching >= maxclauses10 ) 
      errorwithoutline( "clauses is full and cannot be extended from here" ); 
#endif
   *tree = freeclauses;
   clauses [ freeclauses ++ ] = DT_SUCC;
   clauses [ freeclauses ++ ] = DT_NIL; 
   for( i = 0; i < freematching; i ++ ) 
      clauses [ freeclauses ++ ] = matching [i]. var; 
   clauses [ freeclauses ++ ] = clause; 
   clauses [ freeclauses ++ ] = literal; 
   clauses [ freeclauses ++ ] = term;
}


/* delete a term from a discrimimation tree: */      
 

static void deletefromtree2( tree, clause, literal, term, termi, end ) 
int *tree;
int clause;
int literal;
int term;
int termi; 
int end; 
{
   int f;
   int i;

   if( termi < end )
   {
      f = termspace [ termi ++ ];
      if( f < 0 ) 
      {
         for( i = 0; i < freematching; i ++ )
         {
            if( matching [i]. var == f )
               goto varfound;
         }
         if( freematching >= MAXMATCHING - 3 ) 
            errorwithoutline( "matching is full" ); 

         matching [ freematching ++ ]. var = f;
      varfound: 
         f = -i - 1;
      }
      while( *tree != DT_NIL ) 
      {
         if( clauses [ *tree + 1 ] == f )
         {
            deletefromtree2( &clauses [ *tree + 2 ], clause, literal,
                             term, termi, end );
            if( clauses [ *tree + 2 ] == DT_NIL ) 
            {
               *tree = clauses [ *tree + 3 ];
            }
            return;
         }
         tree = &clauses [ *tree + 3 ];
      }

      /* This point should not be reached. Make it known! */
      internalproblem( "deleted non-existing literal from tree" );
   } 
   else 
   {
      /* look for the term in the list of solutions. */

      while( *tree != DT_NIL ) 
      {
         i = *tree + 2;
         while( clauses [i] < 0 ) 
            i ++ ;
         if( clauses [i+0] == clause &&
             clauses [i+1] == literal &&
             clauses [i+2] == term )
         {
            *tree = clauses [ *tree + 1 ];
            return;
         }
         tree = &clauses [ *tree + 1 ]; 
      }
      /* this point should not be reached. */
      internalproblem( "deleted non-existing literal" );
   }
}


void deletefromtree( tree, clause, literal, term )
int *tree;
int clause;
int literal;
int term;
{
   if( termends [ term ] - term > 10000 )
      return;

   freematching = 0;
   deletefromtree2( tree, clause, literal, 
                    term, term, termends [ term ] );
}


/***********************************************************/
/* Retrieve a single term.                                 */
/***********************************************************/



static void retrievefromtree2( tree, term, end, successfunc )
int tree;
int term;
int end;
void successfunc( int, int, int ); 
{
   int f; 
   int i;

   if( term == end ) 
   {
      /* countsolu( tree ); */

      while( tree != DT_NIL )
      {
         /* enumerate solutions. */

         for( i = 0; i < freematching; i ++ )
         {
            matching [i]. var = clauses [ tree + 2 + i ];
         }
         if( clauses [ tree + 2 + i ] < 0 )
         {
            internalproblem( "clauses [ tree + 2 + i ] < 0" );
            exit(0);
         }
         
         successfunc( clauses [ tree + 2 + i ],
                      clauses [ tree + 3 + i ],
                      clauses [ tree + 4 + i ] );  
         tree = clauses [ tree + 1 ];
      }
      return;
   }
   else
   {
      /* countaltern( tree ); */
 
      while( tree != DT_NIL ) 
      {
         f = clauses [ tree + 1 ];
         if( f >= 0 )
         {
            if( f == termspace [ term ] ) 
            {
               retrievefromtree2( clauses [ tree + 2 ], term + 1, 
                  end, successfunc );
            }
         }
         else
         {
            f = -f - 1;
            if( f > freematching ) 
            {
               internalproblem( "f > freematching" );
               exit(0);
            }
            if( f < freematching )
            {
               if( termequal( matching [f]. term, term )) 
                  retrievefromtree2( clauses [ tree + 2 ], termends [ term ],
                                     end, successfunc );
            }
            else
            {
               matching [ freematching ++ ]. term = term;
               retrievefromtree2( clauses [ tree + 2 ], termends [ term ],
                                  end, successfunc ); 
               freematching -- ;
            }
         }
         tree = clauses [ tree + 3 ];
      }
   }
}


/* Retrieve a term that can be matched into term.  */

void retrievefromtree( tree, term, successfunc )
int tree;
int term;
void successfunc( int, int, int );
{
   freematching = 0; 
   retrievefromtree2( tree, term, termends [ term ], successfunc );
   if( freematching != 0 ) 
   {
      internalproblem( "freematching != 0" );
      exit(0);
   }
}


/* Terms to be retrieved from tree:   */

static int termlist [4];
static int nrterms;


static void retrievelistfromtree( tree, termnr, indinterm, endofterm,
          successfunc )
int tree;
int termnr;
int indinterm;
int endofterm;
void successfunc( int, int, int ); 
{
   int f; 
   int i;

#if 0
   printstring( &outputfile, "\n\nRetrievelistfromtree:\n");
   printterm( &outputfile, termlist [0] );
   printstring( &outputfile, " ");
   printterm( &outputfile, termlist [1] );
   printf("%d, %d %d\n", termnr, indinterm - termlist [ termnr ], 
                                 endofterm - termlist [ termnr ] );
#endif
   if( indinterm == endofterm )
   {
      termnr ++ ;
      if( termnr >= nrterms )
      {
         while( tree != DT_NIL )
         {
            /* enumerate solutions. */

            for( i = 0; i < freematching; i ++ )
            {
               matching [i]. var = clauses [ tree + 2 + i ];
            }
            if( clauses [ tree + 2 + i ] < 0 ) 
            {
               internalproblem( "clauses [ tree + 2 + i ] < 0" );
               exit(0);
            }
         
            successfunc( clauses [ tree + 2 + i ],
                         clauses [ tree + 3 + i ],
                         clauses [ tree + 4 + i ] );  
            tree = clauses [ tree + 1 ];
         }
         return;
      }
      else
      {
         indinterm = termlist [ termnr ];
         endofterm = termends [ indinterm ];
      }
   }

   while( tree != DT_NIL ) 
   {
      f = clauses [ tree + 1 ];
      if( f >= 0 )
      {
         if( f == termspace [ indinterm ] ) 
         {
            retrievelistfromtree( clauses [ tree + 2 ], termnr, 
               indinterm + 1, endofterm, successfunc );
         }
      }
      else
      {
         f = -f - 1;
         if( f > freematching ) 
         {
            internalproblem( "f > freematching" );
            exit(0);
         }
         if( f < freematching )
         {
            if( termequal( matching [f]. term, indinterm )) 
               retrievelistfromtree( clauses [ tree + 2 ], termnr, 
                  termends [ indinterm ], endofterm, successfunc );
         }
         else
         {
            matching [ freematching ++ ]. term = indinterm;
            retrievelistfromtree( clauses [ tree + 2 ], termnr, 
               termends [ indinterm ], endofterm, successfunc ); 
            freematching -- ;
         }
      }
      tree = clauses [ tree + 3 ];
   }
}


/* Retrieve a literal, ignoring the difference between = and ==>,  */
/* and allowing them to be commutative.                            */


void retrievefromtreeeqrewr( tree, term, successfunc )
int tree;
int term;
void successfunc( int, int, int );
{
   freematching = 0;

   if( termspace [ term ] == OP_NOT )
   {
      while( tree != DT_NIL && clauses [ tree + 1 ] != OP_NOT )
          tree = clauses [ tree + 3 ];
      if( tree == DT_NIL )
         return;  
      tree = clauses [ tree + 2 ];
      term ++ ;
   }

   if( termspace [ term ] == OP_EQUAL || termspace [ term ] == OP_REWRITES )
   {

      term ++ ;
      while( tree != DT_NIL ) 
      {
         if( clauses [ tree + 1 ] == OP_EQUAL || 
             clauses [ tree + 1 ] == OP_REWRITES )
         {
            termlist [0] = term;
            termlist [1] = termends [ term ];
            nrterms = 2;
            retrievelistfromtree( clauses [ tree + 2 ], 0,
                                  termlist [0], termends [ termlist [0]], 
                                  successfunc );

            termlist [0] = termends [ term ];
            termlist [1] = term;
            nrterms = 2; 
            retrievelistfromtree( clauses [ tree + 2 ], 0,
                                  termlist [0], termends [ termlist [0]], 
                                  successfunc );
         }
         tree = clauses [ tree + 3 ];
      }
   }
   else
      retrievefromtree2( tree, term, termends [ term ], successfunc );

   if( freematching != 0 ) 
   {
      internalproblem( "freematching != 0" );
      exit(0);
   }
}


#if 0

/* For debugging retrievefrom tree. */ 

void printtreeresult( clause, literal, term )
int clause;
int literal;
int term;
{
   printstring( &outputfile, "Found a solution:\n\n" );
   printclause( &outputfile, clause );
   printstring( &outputfile, "Literal:  " ); 
   printterm( &outputfile, clauses [ literal ] );
   printstring( &outputfile, ",   term:   " ); 
   printterm( &outputfile, term );
   printstring( &outputfile, "\n" ); 
   printmatching( &outputfile );
}

#endif 
